# Copyright 2016 <PUT YOUR NAME/COMPANY HERE>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Drop v1 tables

Revision ID: e6417a8b114d
Create Date: 2016-08-23 12:48:46.985939

"""

from alembic import op

from neutron.db import migration


revision = 'e6417a8b114d'
down_revision = '4b4dc6d5d843'


# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.NEWTON]


def upgrade():
    op.drop_table('nsxv_edge_pool_mappings')
    op.drop_table('nsxv_edge_vip_mappings')
    op.drop_table('nsxv_edge_monitor_mappings')
    op.drop_table('members')
    op.drop_table('poolstatisticss')
    op.drop_table('poolloadbalanceragentbindings')
    op.drop_table('poolmonitorassociations')
    op.drop_table('pools')
    op.drop_table('sessionpersistences')
    op.drop_table('vips')
    op.drop_table('healthmonitors')



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import socket

import netaddr
from neutron.agent.linux import external_process
from neutron.agent.linux import ip_lib
from neutron_lib import constants
from neutron_lib import exceptions
from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import fileutils

from neutron_lbaas._i18n import _
from neutron_lbaas.agent import agent_device_driver
from neutron_lbaas.drivers.haproxy import jinja_cfg
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import data_models

LOG = logging.getLogger(__name__)
NS_PREFIX = 'qlbaas-'
STATS_TYPE_BACKEND_REQUEST = 2
STATS_TYPE_BACKEND_RESPONSE = '1'
STATS_TYPE_SERVER_REQUEST = 4
STATS_TYPE_SERVER_RESPONSE = '2'
DRIVER_NAME = 'haproxy_ns'
HAPROXY_SERVICE_NAME = 'lbaas-ns-haproxy'

STATE_PATH_V2_APPEND = 'v2'

STATE_PATH_DEFAULT = '$state_path/lbaas'
USER_GROUP_DEFAULT = 'nogroup'
OPTS = [
    cfg.StrOpt(
        'loadbalancer_state_path',
        default=STATE_PATH_DEFAULT,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Location to store config and state files'),
        deprecated_opts=[cfg.DeprecatedOpt('loadbalancer_state_path',
                                           group='DEFAULT')],
    ),
    cfg.StrOpt(
        'user_group',
        default=USER_GROUP_DEFAULT,
        help=_('The user group'),
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        deprecated_opts=[cfg.DeprecatedOpt('user_group', group='DEFAULT')],
    ),
    cfg.IntOpt(
        'send_gratuitous_arp',
        default=3,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('When delete and re-add the same vip, send this many '
               'gratuitous ARPs to flush the ARP cache in the Router. '
               'Set it below or equal to 0 to disable this feature.'),
    )
]
cfg.CONF.register_opts(OPTS, 'haproxy')


def get_ns_name(namespace_id):
    return NS_PREFIX + namespace_id


class HaproxyNSDriver(agent_device_driver.AgentDeviceDriver):

    def __init__(self, conf, plugin_rpc, process_monitor):
        super(HaproxyNSDriver, self).__init__(conf, plugin_rpc,
                                              process_monitor)
        self.state_path = conf.haproxy.loadbalancer_state_path
        self.state_path = os.path.join(
            self.conf.haproxy.loadbalancer_state_path, STATE_PATH_V2_APPEND)
        try:
            vif_driver_class = runtime.load_class_by_alias_or_classname(
                'neutron.interface_drivers',
                conf.interface_driver)
        except ImportError:
            with excutils.save_and_reraise_exception():
                msg = (_('Error importing interface driver: %s')
                       % conf.interface_driver)
                LOG.error(msg)

        self.vif_driver = vif_driver_class(conf)
        self.deployed_loadbalancers = {}
        self._loadbalancer = LoadBalancerManager(self)
        self._listener = ListenerManager(self)
        self._pool = PoolManager(self)
        self._member = MemberManager(self)
        self._healthmonitor = HealthMonitorManager(self)

    @property
    def loadbalancer(self):
        return self._loadbalancer

    @property
    def listener(self):
        return self._listener

    @property
    def pool(self):
        return self._pool

    @property
    def member(self):
        return self._member

    @property
    def healthmonitor(self):
        return self._healthmonitor

    def get_name(self):
        return DRIVER_NAME

    @runtime.synchronized('haproxy-driver')
    def undeploy_instance(self, loadbalancer_id, **kwargs):
        cleanup_namespace = kwargs.get('cleanup_namespace', False)
        delete_namespace = kwargs.get('delete_namespace', False)
        namespace = get_ns_name(loadbalancer_id)
        pid_data = self._get_state_file_path(loadbalancer_id, 'haproxy.pid')
        pid_path = os.path.split(pid_data)[0]
        self.process_monitor.unregister(uuid=loadbalancer_id,
                                        service_name=HAPROXY_SERVICE_NAME)
        pm = external_process.ProcessManager(uuid=loadbalancer_id,
                                             namespace=namespace,
                                             service=HAPROXY_SERVICE_NAME,
                                             conf=self.conf,
                                             pids_path=pid_path,
                                             pid_file=pid_data)
        pm.disable()
        # unplug the ports
        if loadbalancer_id in self.deployed_loadbalancers:
            self._unplug(namespace,
                         self.deployed_loadbalancers[loadbalancer_id].vip_port)

        # delete all devices from namespace
        # used when deleting orphans and port is not known for a loadbalancer
        if cleanup_namespace:
            ns = ip_lib.IPWrapper(namespace=namespace)
            for device in ns.get_devices(exclude_loopback=True):
                self.vif_driver.unplug(device.name, namespace=namespace)

        # remove the configuration directory
        conf_dir = os.path.dirname(
            self._get_state_file_path(loadbalancer_id, ''))
        if os.path.isdir(conf_dir):
            shutil.rmtree(conf_dir)

        if delete_namespace:
            ns = ip_lib.IPWrapper(namespace=namespace)
            ns.garbage_collect_namespace()

    def remove_orphans(self, known_loadbalancer_ids):
        if not os.path.exists(self.state_path):
            return

        orphans = (lb_id for lb_id in os.listdir(self.state_path)
                   if lb_id not in known_loadbalancer_ids)
        for lb_id in orphans:
            if self.exists(lb_id):
                self.undeploy_instance(lb_id, cleanup_namespace=True)

    def get_stats(self, loadbalancer_id):
        socket_path = self._get_state_file_path(loadbalancer_id,
                                                'haproxy_stats.sock', False)
        if os.path.exists(socket_path):
            parsed_stats = self._get_stats_from_socket(
                socket_path,
                entity_type=(STATS_TYPE_BACKEND_REQUEST |
                             STATS_TYPE_SERVER_REQUEST))
            lb_stats = self._get_backend_stats(parsed_stats)
            lb_stats['members'] = self._get_servers_stats(parsed_stats)
            return lb_stats
        else:
            lb_config = self.plugin_rpc.get_loadbalancer(loadbalancer_id)
            loadbalancer = data_models.LoadBalancer.from_dict(lb_config)
            if self._is_active(loadbalancer):
                LOG.warning('Stats socket not found for loadbalancer %s',
                            loadbalancer_id)
            else:
                LOG.debug('Stats socket not found for loadbalancer %s,'
                          ' but loadbalancer has no VIP in state UP.'
                          ' Perhaps the lbaas-listener is not yet created?',
                          loadbalancer_id)
            return {}

    @runtime.synchronized('haproxy-driver')
    def deploy_instance(self, loadbalancer):
        """Deploys loadbalancer if necessary

        :returns: True if loadbalancer was deployed, False otherwise
        """
        if not self.deployable(loadbalancer):
            LOG.info("Loadbalancer %s is not deployable." %
                     loadbalancer.id)
            return False

        if self.exists(loadbalancer.id):
            self.update(loadbalancer)
        else:
            self.create(loadbalancer)
        return True

    def update(self, loadbalancer):
        pid_path = self._get_state_file_path(loadbalancer.id, 'haproxy.pid')
        extra_args = ['-sf']
        extra_args.extend(p.strip() for p in open(pid_path, 'r'))
        self._spawn(loadbalancer, extra_args)

    def exists(self, loadbalancer_id):
        namespace = get_ns_name(loadbalancer_id)
        root_ns = ip_lib.IPWrapper()

        socket_path = self._get_state_file_path(
            loadbalancer_id, 'haproxy_stats.sock', False)
        if root_ns.netns.exists(namespace) and os.path.exists(socket_path):
            try:
                s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                s.connect(socket_path)
                return True
            except socket.error:
                pass
        return False

    def create(self, loadbalancer):
        namespace = get_ns_name(loadbalancer.id)

        self._plug(namespace, loadbalancer.vip_port, loadbalancer.vip_address)
        self._spawn(loadbalancer)

    def deployable(self, loadbalancer):
        """Returns True if loadbalancer is active and has active listeners."""
        if not loadbalancer:
            return False
        acceptable_listeners = [
            listener for listener in loadbalancer.listeners
            if (listener.provisioning_status != constants.PENDING_DELETE and
                listener.admin_state_up)]
        return (bool(acceptable_listeners) and loadbalancer.admin_state_up and
                loadbalancer.provisioning_status != constants.PENDING_DELETE)

    def _get_stats_from_socket(self, socket_path, entity_type):
        try:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect(socket_path)
            s.send(('show stat -1 %s -1\n' % entity_type).encode('utf-8'))
            raw_stats = ''
            chunk_size = 1024
            while True:
                chunk = s.recv(chunk_size).decode('utf-8')
                raw_stats += chunk
                if len(chunk) < chunk_size:
                    break

            return self._parse_stats(raw_stats)
        except socket.error as e:
            LOG.warning('Error while connecting to stats socket: %s', e)
            return {}

    def _parse_stats(self, raw_stats):
        stat_lines = raw_stats.splitlines()
        if len(stat_lines) < 2:
            return []
        formatted_stat_names = stat_lines[0].lstrip('# ').rstrip(',')
        stat_names = formatted_stat_names.split(',')
        res_stats = []
        for raw_values in stat_lines[1:]:
            if not raw_values:
                continue
            stat_values = [value.strip()
                           for value in raw_values.rstrip(',').split(',')]
            res_stats.append(dict(zip(stat_names, stat_values)))

        return res_stats

    def _get_backend_stats(self, parsed_stats):
        for stats in parsed_stats:
            if stats.get('type') == STATS_TYPE_BACKEND_RESPONSE:
                unified_stats = dict((k, stats.get(v, ''))
                                     for k, v in jinja_cfg.STATS_MAP.items())
                return unified_stats

        return {}

    def _get_servers_stats(self, parsed_stats):
        res = {}
        for stats in parsed_stats:
            if stats.get('type') == STATS_TYPE_SERVER_RESPONSE:
                res[stats['svname']] = {
                    lb_const.OPERATING_STATUS: (lb_const.OFFLINE
                                            if stats['status'] == 'DOWN'
                                            else lb_const.ONLINE),
                    lb_const.STATS_HEALTH: stats['check_status'],
                    lb_const.STATS_FAILED_CHECKS: stats['chkfail']
                }
        return res

    def _get_state_file_path(self, loadbalancer_id, kind,
                             ensure_state_dir=True):
        """Returns the file name for a given kind of config file."""
        confs_dir = os.path.abspath(os.path.normpath(self.state_path))
        conf_dir = os.path.join(confs_dir, loadbalancer_id)
        if ensure_state_dir:
            fileutils.ensure_tree(conf_dir, 0o755)
        return os.path.join(conf_dir, kind)

    def _plug(self, namespace, port, vip_address, reuse_existing=True):
        self.plugin_rpc.plug_vip_port(port.id)

        interface_name = self.vif_driver.get_device_name(port)

        if ip_lib.device_exists(interface_name,
                                namespace=namespace):
            if not reuse_existing:
                raise exceptions.PreexistingDeviceFailure(
                    dev_name=interface_name
                )
        else:
            self.vif_driver.plug(
                port.network_id,
                port.id,
                interface_name,
                port.mac_address,
                namespace=namespace,
                mtu=port.network.mtu
            )

        cidrs = [
            '%s/%s' % (ip.ip_address,
                       netaddr.IPNetwork(ip.subnet.cidr).prefixlen)
            for ip in port.fixed_ips
        ]
        self.vif_driver.init_l3(interface_name, cidrs, namespace=namespace)

        # Haproxy socket binding to IPv6 VIP address will fail if this address
        # is not yet ready(i.e tentative address).
        if netaddr.IPAddress(vip_address).version == 6:
            device = ip_lib.IPDevice(interface_name, namespace=namespace)
            device.addr.wait_until_address_ready(vip_address)

        # Add subnet host routes
        host_routes = port.fixed_ips[0].subnet.host_routes
        for host_route in host_routes:
            if host_route.destination != "0.0.0.0/0":
                cmd = ['route', 'add', '-net', host_route.destination,
                       'gw', host_route.nexthop]
                ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
                ip_wrapper.netns.execute(cmd, check_exit_code=False)

        gw_ip = port.fixed_ips[0].subnet.gateway_ip

        if not gw_ip:
            for host_route in host_routes:
                if host_route.destination == "0.0.0.0/0":
                    gw_ip = host_route.nexthop
                    break
        else:
            cmd = ['route', 'add', 'default', 'gw', gw_ip]
            ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
            ip_wrapper.netns.execute(cmd, check_exit_code=False)
            # When delete and re-add the same vip, we need to
            # send gratuitous ARP to flush the ARP cache in the Router.
            gratuitous_arp = self.conf.haproxy.send_gratuitous_arp
            if gratuitous_arp > 0:
                for ip in port.fixed_ips:
                    cmd_arping = ['arping', '-U',
                                  '-I', interface_name,
                                  '-c', gratuitous_arp,
                                  ip.ip_address]
                    ip_wrapper.netns.execute(cmd_arping, check_exit_code=False)

    def _unplug(self, namespace, port):
        self.plugin_rpc.unplug_vip_port(port.id)
        interface_name = self.vif_driver.get_device_name(port)
        self.vif_driver.unplug(interface_name, namespace=namespace)

    def _spawn(self, loadbalancer, extra_cmd_args=()):
        conf_path = self._get_state_file_path(loadbalancer.id,
                                              'haproxy.conf')
        sock_path = self._get_state_file_path(loadbalancer.id,
                                              'haproxy_stats.sock')
        user_group = self.conf.haproxy.user_group
        haproxy_base_dir = self._get_state_file_path(loadbalancer.id, '')
        jinja_cfg.save_config(conf_path,
                              loadbalancer,
                              sock_path,
                              user_group,
                              haproxy_base_dir)

        def callback(pid_path):
            cmd = ['haproxy', '-f', conf_path, '-p', pid_path]
            cmd.extend(extra_cmd_args)
            return cmd

        pid_data = self._get_state_file_path(loadbalancer.id, 'haproxy.pid')
        pid_path = os.path.split(pid_data)[0]
        namespace = get_ns_name(loadbalancer.id)
        pm = external_process.ProcessManager(
            uuid=loadbalancer.id,
            default_cmd_callback=callback,
            namespace=namespace,
            service=HAPROXY_SERVICE_NAME,
            conf=self.conf,
            pids_path=pid_path,
            pid_file=pid_data,
            custom_reload_callback=callback if extra_cmd_args else None)
        if pm.active:
            pm.reload_cfg()
        else:
            pm.enable()
        self.process_monitor.register(uuid=loadbalancer.id,
                                      service_name=HAPROXY_SERVICE_NAME,
                                      monitored_process=pm)
        # remember deployed loadbalancer id
        self.deployed_loadbalancers[loadbalancer.id] = loadbalancer

    def _is_active(self, loadbalancer):
        # haproxy will be unable to start without any active vip
        if (len(loadbalancer.listeners) == 0 or
                loadbalancer.vip_port.status == constants.DOWN or
                not loadbalancer.vip_port.admin_state_up or
                loadbalancer.operating_status != lb_const.ONLINE):
            return False
        return True


class LoadBalancerManager(agent_device_driver.BaseLoadBalancerManager):

    def refresh(self, loadbalancer):
        loadbalancer_dict = self.driver.plugin_rpc.get_loadbalancer(
            loadbalancer.id)
        loadbalancer = data_models.LoadBalancer.from_dict(loadbalancer_dict)
        if (not self.driver.deploy_instance(loadbalancer) and
                self.driver.exists(loadbalancer.id)):
            self.driver.undeploy_instance(loadbalancer.id)

    def delete(self, loadbalancer):
        if self.driver.exists(loadbalancer.id):
            self.driver.undeploy_instance(loadbalancer.id,
                                          delete_namespace=True)

    def create(self, loadbalancer):
        # loadbalancer has no listeners then do nothing because haproxy will
        # not start without a tcp port.  Consider this successful anyway.
        if not loadbalancer.listeners:
            return
        self.refresh(loadbalancer)

    def get_stats(self, loadbalancer_id):
        stats = self.driver.get_stats(loadbalancer_id)

        # NOTE(cgoncalves): haproxy stats include member status which maps to
        # MemberV2.operating_status. Take the opportunty to update member
        # operating status.
        members = stats.get('members')
        if members:
            for member_id, value in members.items():
                if lb_const.OPERATING_STATUS in value:
                    self.driver.plugin_rpc.update_status(
                        'member', member_id,
                        operating_status=value[lb_const.OPERATING_STATUS])

        return stats

    def update(self, old_loadbalancer, loadbalancer):
        self.refresh(loadbalancer)


class ListenerManager(agent_device_driver.BaseListenerManager):

    def _remove_listener(self, loadbalancer, listener_id):
        index_to_remove = None
        for index, listener in enumerate(loadbalancer.listeners):
            if listener.id == listener_id:
                index_to_remove = index
        loadbalancer.listeners.pop(index_to_remove)

    def update(self, old_listener, new_listener):
        self.driver.loadbalancer.refresh(new_listener.loadbalancer)

    def create(self, listener):
        self.driver.loadbalancer.refresh(listener.loadbalancer)

    def delete(self, listener):
        loadbalancer = listener.loadbalancer
        self._remove_listener(loadbalancer, listener.id)
        if len(loadbalancer.listeners) > 0:
            self.driver.loadbalancer.refresh(loadbalancer)
        else:
            # undeploy instance because haproxy will throw error if port is
            # missing in frontend
            self.driver.undeploy_instance(loadbalancer.id,
                                          delete_namespace=True)


class PoolManager(agent_device_driver.BasePoolManager):

    def update(self, old_pool, new_pool):
        self.driver.loadbalancer.refresh(new_pool.loadbalancer)

    def create(self, pool):
        self.driver.loadbalancer.refresh(pool.loadbalancer)

    def delete(self, pool):
        loadbalancer = pool.loadbalancer
        for l in loadbalancer.listeners:
            if l.default_pool == pool:
                l.default_pool = None
        # TODO(sbalukoff): Will need to do this or L7Policies as well
        # just refresh because haproxy is fine if only frontend is listed
        self.driver.loadbalancer.refresh(loadbalancer)


class MemberManager(agent_device_driver.BaseMemberManager):

    def _remove_member(self, pool, member_id):
        index_to_remove = None
        for index, member in enumerate(pool.members):
            if member.id == member_id:
                index_to_remove = index
        pool.members.pop(index_to_remove)

    def update(self, old_member, new_member):
        self.driver.loadbalancer.refresh(new_member.pool.loadbalancer)

    def create(self, member):
        self.driver.loadbalancer.refresh(member.pool.loadbalancer)

    def delete(self, member):
        self._remove_member(member.pool, member.id)
        self.driver.loadbalancer.refresh(member.pool.loadbalancer)


class HealthMonitorManager(agent_device_driver.BaseHealthMonitorManager):

    def update(self, old_hm, new_hm):
        self.driver.loadbalancer.refresh(new_hm.pool.loadbalancer)

    def create(self, hm):
        self.driver.loadbalancer.refresh(hm.pool.loadbalancer)

    def delete(self, hm):
        hm.pool.healthmonitor = None
        self.driver.loadbalancer.refresh(hm.pool.loadbalancer)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from neutron_lbaas.common.cert_manager import cert_manager
from neutron_lbaas.common.tls_utils import cert_parser
from neutron_lbaas.drivers.haproxy import jinja_cfg
from neutron_lbaas.services.loadbalancer import constants
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests.unit.drivers.haproxy.\
    sample_configs import sample_configs


class TestHaproxyCfg(base.BaseTestCase):
    def test_save_config(self):
        with mock.patch('neutron_lbaas.drivers.haproxy.'
                        'jinja_cfg.render_loadbalancer_obj') as r_t, \
                mock.patch('neutron_lib.utils.file.replace_file') as replace:
            r_t.return_value = 'fake_rendered_template'
            lb = mock.Mock()
            jinja_cfg.save_config('test_conf_path', lb, 'test_sock_path',
                                  'nogroup',
                                  'fake_state_path')
            r_t.assert_called_once_with(lb,
                                        'nogroup',
                                        'test_sock_path',
                                        'fake_state_path')
            replace.assert_called_once_with('test_conf_path',
                                            'fake_rendered_template')

    def test_get_template(self):
        template = jinja_cfg._get_template()
        self.assertEqual('haproxy.loadbalancer.j2', template.name)

    def test_render_template_tls_termination(self):
        lb = sample_configs.sample_loadbalancer_tuple(
            proto=constants.PROTOCOL_TERMINATED_HTTPS, tls=True, sni=True)

        fe = ("frontend sample_listener_id_1\n"
              "    option httplog\n"
              "    redirect scheme https if !{ ssl_fc }\n"
              "    maxconn 98\n"
              "    option forwardfor\n"
              "    bind 10.0.0.2:443"
              " ssl crt /v2/sample_listener_id_1/fakeCNM.pem"
              " crt /v2/sample_listener_id_1\n"
              "    mode http\n"
              "    default_backend sample_pool_id_1\n\n")
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    timeout check 31s\n"
              "    option httpchk GET /index.html\n"
              "    http-check expect rstatus %s\n"
              "    server sample_member_id_1 10.0.0.99:82"
              " weight 13 check inter 30s fall 3 cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82"
              " weight 13 check inter 30s fall 3 cookie "
              "sample_member_id_2\n\n"
              % sample_configs.PIPED_CODES)
        with mock.patch('os.makedirs'):
            with mock.patch('os.listdir'):
                with mock.patch.object(jinja_cfg, 'file_utils'):
                    with mock.patch.object(
                            jinja_cfg, '_process_tls_certificates') as crt:
                        crt.return_value = {
                            'tls_cert': lb.listeners[0]
                            .default_tls_container,
                            'sni_certs': [lb.listeners[0]
                                          .sni_containers[0].tls_container]}
                        rendered_obj = jinja_cfg.render_loadbalancer_obj(
                            lb, 'nogroup',
                            '/sock_path',
                            '/v2')
                        self.assertEqual(
                            sample_configs.sample_base_expected_config(
                                frontend=fe, backend=be),
                            rendered_obj)

    def test_render_template_tls_termination_no_sni(self):
        lb = sample_configs.sample_loadbalancer_tuple(
            proto=constants.PROTOCOL_TERMINATED_HTTPS, tls=True)

        fe = ("frontend sample_listener_id_1\n"
              "    option httplog\n"
              "    redirect scheme https if !{ ssl_fc }\n"
              "    maxconn 98\n"
              "    option forwardfor\n"
              "    bind 10.0.0.2:443"
              " ssl crt /v2/sample_listener_id_1/fakeCNM.pem\n"
              "    mode http\n"
              "    default_backend sample_pool_id_1\n\n")
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    timeout check 31s\n"
              "    option httpchk GET /index.html\n"
              "    http-check expect rstatus %s\n"
              "    server sample_member_id_1 10.0.0.99:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_2\n\n"
              % sample_configs.PIPED_CODES)
        with mock.patch('os.makedirs'):
            with mock.patch('neutron_lib.utils.file.replace_file'):
                with mock.patch('os.listdir'):
                    with mock.patch.object(jinja_cfg, 'file_utils'):
                        with mock.patch.object(
                                jinja_cfg, '_process_tls_certificates') as crt:
                            crt.return_value = {
                                'tls_cert': lb.listeners[0]
                                .default_tls_container,
                                'sni_certs': []}
                            rendered_obj = jinja_cfg.render_loadbalancer_obj(
                                lb, 'nogroup',
                                '/sock_path',
                                '/v2')
                            self.assertEqual(
                                sample_configs.sample_base_expected_config(
                                    frontend=fe, backend=be),
                                rendered_obj)

    def test_render_template_http(self):
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    timeout check 31s\n"
              "    option httpchk GET /index.html\n"
              "    http-check expect rstatus %s\n"
              "    server sample_member_id_1 10.0.0.99:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_2\n\n"
              % sample_configs.PIPED_CODES)
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(
            sample_configs.sample_base_expected_config(backend=be),
            rendered_obj)

    def test_render_template_https(self):
        fe = ("frontend sample_listener_id_1\n"
              "    option tcplog\n"
              "    maxconn 98\n"
              "    bind 10.0.0.2:443\n"
              "    mode tcp\n"
              "    default_backend sample_pool_id_1\n\n")
        be = ("backend sample_pool_id_1\n"
              "    mode tcp\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    timeout check 31s\n"
              "    option httpchk GET /index.html\n"
              "    http-check expect rstatus %s\n"
              "    option ssl-hello-chk\n"
              "    server sample_member_id_1 10.0.0.99:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82 "
              "weight 13 check inter 30s fall 3 cookie sample_member_id_2\n\n"
              % sample_configs.PIPED_CODES)
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(proto='HTTPS'),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(sample_configs.sample_base_expected_config(
            frontend=fe, backend=be), rendered_obj)

    def test_render_template_no_monitor_http(self):
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    server sample_member_id_1 10.0.0.99:82 weight 13 "
              "cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82 weight 13 "
              "cookie sample_member_id_2\n\n")
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(
                proto=constants.PROTOCOL_HTTP, monitor=False),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(sample_configs.sample_base_expected_config(
            backend=be), rendered_obj)

    def test_render_template_no_monitor_https(self):
        fe = ("frontend sample_listener_id_1\n"
              "    option tcplog\n"
              "    maxconn 98\n"
              "    bind 10.0.0.2:443\n"
              "    mode tcp\n"
              "    default_backend sample_pool_id_1\n\n")
        be = ("backend sample_pool_id_1\n"
              "    mode tcp\n"
              "    balance roundrobin\n"
              "    cookie SRV insert indirect nocache\n"
              "    server sample_member_id_1 10.0.0.99:82 weight 13 "
              "cookie sample_member_id_1\n"
              "    server sample_member_id_2 10.0.0.98:82 weight 13 "
              "cookie sample_member_id_2\n\n")
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(
                proto='HTTPS', monitor=False),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(sample_configs.sample_base_expected_config(
            frontend=fe, backend=be), rendered_obj)

    def test_render_template_no_persistence_https(self):
        fe = ("frontend sample_listener_id_1\n"
              "    option tcplog\n"
              "    maxconn 98\n"
              "    bind 10.0.0.2:443\n"
              "    mode tcp\n"
              "    default_backend sample_pool_id_1\n\n")
        be = ("backend sample_pool_id_1\n"
              "    mode tcp\n"
              "    balance roundrobin\n"
              "    server sample_member_id_1 10.0.0.99:82 weight 13\n"
              "    server sample_member_id_2 10.0.0.98:82 weight 13\n\n")
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(
                proto='HTTPS', monitor=False, persistence=False),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(sample_configs.sample_base_expected_config(
            frontend=fe, backend=be), rendered_obj)

    def test_render_template_no_persistence_http(self):
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    server sample_member_id_1 10.0.0.99:82 weight 13\n"
              "    server sample_member_id_2 10.0.0.98:82 weight 13\n\n")
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(
                proto=constants.PROTOCOL_HTTP, monitor=False,
                persistence=False),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(sample_configs.sample_base_expected_config(
            backend=be), rendered_obj)

    def test_render_template_sourceip_persistence(self):
        be = ("backend sample_pool_id_1\n"
              "    mode http\n"
              "    balance roundrobin\n"
              "    stick-table type ip size 10k\n"
              "    stick on src\n"
              "    timeout check 31s\n"
              "    option httpchk GET /index.html\n"
              "    http-check expect rstatus %s\n"
              "    server sample_member_id_1 10.0.0.99:82 "
              "weight 13 check inter 30s fall 3\n"
              "    server sample_member_id_2 10.0.0.98:82 "
              "weight 13 check inter 30s fall 3\n\n"
              % sample_configs.PIPED_CODES)
        rendered_obj = jinja_cfg.render_loadbalancer_obj(
            sample_configs.sample_loadbalancer_tuple(
                persistence_type='SOURCE_IP'),
            'nogroup', '/sock_path', '/v2')
        self.assertEqual(
            sample_configs.sample_base_expected_config(backend=be),
            rendered_obj)

    def test_render_template_appsession_persistence(self):
        with mock.patch('os.makedirs') as md:
            with mock.patch.object(jinja_cfg, 'file_utils'):
                md.return_value = '/data/dirs/'
                be = ("backend sample_pool_id_1\n"
                      "    mode http\n"
                      "    balance roundrobin\n"
                      "    stick-table type string len 64 size 10k\n"
                      "    stick store-response res.cook(APP_COOKIE)\n"
                      "    stick match req.cook(APP_COOKIE)\n"
                      "    timeout check 31s\n"
                      "    option httpchk GET /index.html\n"
                      "    http-check expect rstatus %s\n"
                      "    server sample_member_id_1 10.0.0.99:82 "
                      "weight 13 check inter 30s fall 3\n"
                      "    server sample_member_id_2 10.0.0.98:82 "
                      "weight 13 check inter 30s fall 3\n\n"
                      % sample_configs.PIPED_CODES)
                rendered_obj = jinja_cfg.render_loadbalancer_obj(
                    sample_configs.sample_loadbalancer_tuple(
                        persistence_type='APP_COOKIE'),
                    'nogroup', '/sock_path',
                    '/v2')
                self.assertEqual(
                    sample_configs.sample_base_expected_config(backend=be),
                    rendered_obj)

    def test_retrieve_crt_path(self):
        with mock.patch('os.makedirs'):
            with mock.patch('os.path.isdir') as isdir:
                with mock.patch.object(jinja_cfg, '_retrieve_crt_path') as rcp:
                    isdir.return_value = True
                    rcp.return_value = '/v2/loadbalancers/lb_id_1/' \
                                       'cont_id_1.pem'
                    ret = jinja_cfg._retrieve_crt_path(
                        '/v2/loadbalancers', 'lb_id_1', 'cont_id_1')
                    self.assertEqual(
                        '/v2/loadbalancers/lb_id_1/cont_id_1.pem', ret)

    def test_store_listener_crt(self):
        l = sample_configs.sample_listener_tuple(tls=True, sni=True)
        with mock.patch('os.makedirs'):
            with mock.patch('neutron_lib.utils.file.replace_file'):
                    ret = jinja_cfg._store_listener_crt(
                        '/v2/loadbalancers', l, l.default_tls_container)
                    self.assertEqual(
                        '/v2/loadbalancers/sample_listener_id_1/fakeCNM.pem',
                        ret)

    def test_process_tls_certificates(self):
        sl = sample_configs.sample_listener_tuple(tls=True, sni=True)
        tls = data_models.TLSContainer(primary_cn='fakeCN',
                                       certificate='imaCert',
                                       private_key='imaPrivateKey',
                                       intermediates=['imainter1',
                                                      'imainter2'])
        cert = mock.Mock(spec=cert_manager.Cert)
        cert.get_private_key.return_value = tls.private_key
        cert.get_certificate.return_value = tls.certificate
        cert.get_intermediates.return_value = tls.intermediates

        with mock.patch.object(jinja_cfg, '_map_cert_tls_container') as map, \
                mock.patch.object(jinja_cfg,
                                  '_store_listener_crt') as store_cert, \
                mock.patch.object(cert_parser,
                                  'get_host_names') as get_host_names, \
                mock.patch.object(jinja_cfg,
                                  'CERT_MANAGER_PLUGIN') as cert_mgr:
            map.return_value = tls
            cert_mgr_mock = mock.Mock(spec=cert_manager.CertManager)
            cert_mgr_mock.get_cert.return_value = cert
            cert_mgr.CertManager.return_value = cert_mgr_mock
            get_host_names.return_value = {'cn': 'fakeCN'}
            jinja_cfg._process_tls_certificates(sl)

            # Ensure get_cert is called three times
            calls_certs = [
                mock.call(sl.default_tls_container.id),
                mock.call('cont_id_2'),
                mock.call('cont_id_3')]
            cert_mgr_mock.get_cert.call_args_list == calls_certs

            # Ensure store_cert is called three times
            calls_ac = [mock.call('/v2/',
                                  'sample_listener_id_1',
                                  tls),
                        mock.call('/v2/',
                                  'sample_listener_id_1',
                                  tls),
                        mock.call('/v2/',
                                  'sample_listener_id_1',
                                  tls)]
            store_cert.call_args_list == calls_ac

    def test_get_primary_cn(self):
        cert = mock.MagicMock()

        with mock.patch.object(cert_parser, 'get_host_names') as cp:
            cp.return_value = {'cn': 'fakeCN'}
            cn = jinja_cfg._get_primary_cn(cert.get_certificate())
            self.assertEqual('fakeCN', cn)

    def test_map_cert_tls_container(self):
        tls = data_models.TLSContainer(primary_cn='fakeCN',
                                       certificate='imaCert',
                                       private_key='imaPrivateKey',
                                       intermediates=['imainter1',
                                                      'imainter2'])
        cert = mock.MagicMock()
        cert.get_private_key.return_value = tls.private_key
        cert.get_certificate.return_value = tls.certificate
        cert.get_intermediates.return_value = tls.intermediates
        cert.get_private_key_passphrase.return_value = 'passphrase'
        with mock.patch.object(cert_parser, 'get_host_names') as cp:
            with mock.patch.object(cert_parser, 'dump_private_key') as dp:
                cp.return_value = {'cn': 'fakeCN'}
                dp.return_value = 'imaPrivateKey'
                self.assertEqual(tls.primary_cn,
                                 jinja_cfg._map_cert_tls_container(
                                     cert).primary_cn)
                self.assertEqual(tls.certificate,
                                 jinja_cfg._map_cert_tls_container(
                                     cert).certificate)
                self.assertEqual(tls.private_key,
                                 jinja_cfg._map_cert_tls_container(
                                     cert).private_key)
                self.assertEqual(tls.intermediates,
                                 jinja_cfg._map_cert_tls_container(
                                     cert).intermediates)

    def test_build_pem(self):
        expected = 'imainter\nimainter2\nimacert\nimakey'
        tls_tupe = sample_configs.sample_tls_container_tuple(
            certificate='imacert', private_key='imakey',
            intermediates=['imainter', 'imainter2'])
        self.assertEqual(expected, jinja_cfg._build_pem(tls_tupe))

    def test_transform_session_persistence(self):
        in_persistence = sample_configs.sample_session_persistence_tuple()
        ret = jinja_cfg._transform_session_persistence(in_persistence)
        self.assertEqual(sample_configs.RET_PERSISTENCE, ret)

    def test_transform_health_monitor(self):
        in_persistence = sample_configs.sample_health_monitor_tuple()
        ret = jinja_cfg._transform_health_monitor(in_persistence)
        self.assertEqual(sample_configs.RET_MONITOR, ret)

    def test_transform_member(self):
        in_member = sample_configs.sample_member_tuple('sample_member_id_1',
                                                       '10.0.0.99')
        ret = jinja_cfg._transform_member(in_member)
        self.assertEqual(sample_configs.RET_MEMBER_1, ret)

    def test_transform_pool(self):
        in_pool = sample_configs.sample_pool_tuple()
        ret = jinja_cfg._transform_pool(in_pool)
        self.assertEqual(sample_configs.RET_POOL, ret)

    def test_transform_pool_admin_state_down(self):
        in_pool = sample_configs.sample_pool_tuple(hm_admin_state=False)
        ret = jinja_cfg._transform_pool(in_pool)
        result = sample_configs.RET_POOL
        result['health_monitor'] = ''
        self.assertEqual(result, ret)

    def test_transform_listener(self):
        in_listener = sample_configs.sample_listener_tuple()
        ret = jinja_cfg._transform_listener(in_listener, '/v2')
        self.assertEqual(sample_configs.RET_LISTENER, ret)

    def test_transform_loadbalancer(self):
        in_lb = sample_configs.sample_loadbalancer_tuple()
        ret = jinja_cfg._transform_loadbalancer(in_lb, '/v2')
        self.assertEqual(sample_configs.RET_LB, ret)

    def test_compute_global_connection_limit(self):
        lts = [
                sample_configs.sample_listener_tuple(connection_limit=None),
                sample_configs.sample_listener_tuple()]
        in_listeners = [jinja_cfg._transform_listener(x, '/v2') for x in lts]
        ret = jinja_cfg._compute_global_connection_limit(in_listeners)
        self.assertEqual(2098, ret)

    def test_include_member(self):
        ret = jinja_cfg._include_member(
            sample_configs.sample_member_tuple('sample_member_id_1',
                                               '10.0.0.99'))
        self.assertTrue(ret)

    def test_include_member_invalid_status(self):
        ret = jinja_cfg._include_member(
            sample_configs.sample_member_tuple('sample_member_id_1',
                                               '10.0.0.99', status='PENDING'))
        self.assertFalse(ret)

    def test_include_member_invalid_admin_state(self):
        ret = jinja_cfg._include_member(
            sample_configs.sample_member_tuple('sample_member_id_1',
                                               '10.0.0.99',
                                               admin_state_up=False))
        self.assertFalse(ret)

    def test_expand_expected_codes(self):
        exp_codes = ''
        self.assertEqual(set([]), jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200'
        self.assertEqual(set(['200']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200, 201'
        self.assertEqual(set(['200', '201']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200, 201,202'
        self.assertEqual(set(['200', '201', '202']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200-202'
        self.assertEqual(set(['200', '201', '202']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200-202, 205'
        self.assertEqual(set(['200', '201', '202', '205']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200, 201-203'
        self.assertEqual(set(['200', '201', '202', '203']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '200, 201-203, 205'
        self.assertEqual(set(['200', '201', '202', '203', '205']),
                         jinja_cfg._expand_expected_codes(exp_codes))
        exp_codes = '201-200, 205'
        self.assertEqual(set(['205']),
                         jinja_cfg._expand_expected_codes(exp_codes))

    def test_render_template_about_option_log(self):
        for proto in constants.LISTENER_SUPPORTED_PROTOCOLS:
            proto_mode = jinja_cfg.PROTOCOL_MAP[proto]
            _rendered_obj = jinja_cfg.render_loadbalancer_obj(
                sample_configs.sample_loadbalancer_tuple(
                    proto=proto, monitor=False),
                'nogroup', '/sock_path', '/v2')
            expected_be = \
                ("backend sample_pool_id_1\n"
                 "    mode %s\n"
                 "    balance roundrobin\n"
                 "    cookie SRV insert indirect nocache\n"
                 "    server sample_member_id_1 10.0.0.99:82 weight 13 "
                 "cookie sample_member_id_1\n"
                 "    server sample_member_id_2 10.0.0.98:82 weight 13 "
                 "cookie sample_member_id_2\n\n") % proto_mode
            self.assertEqual(sample_configs.sample_base_expected_config(
                backend=expected_be, fe_proto=proto), _rendered_obj)



# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import collections

from neutron_lbaas.services.loadbalancer import constants

RET_PERSISTENCE = {
    'type': 'HTTP_COOKIE',
    'cookie_name': 'HTTP_COOKIE'}

HASHSEED_ORDERED_CODES = list({'404', '405', '500'})
PIPED_CODES = '|'.join(HASHSEED_ORDERED_CODES)

RET_MONITOR = {
    'id': 'sample_monitor_id_1',
    'type': 'HTTP',
    'delay': 30,
    'timeout': 31,
    'max_retries': 3,
    'http_method': 'GET',
    'url_path': '/index.html',
    'expected_codes': PIPED_CODES,
    'admin_state_up': True}

RET_MEMBER_1 = {
    'id': 'sample_member_id_1',
    'address': '10.0.0.99',
    'protocol_port': 82,
    'weight': 13,
    'subnet_id': '10.0.0.1/24',
    'admin_state_up': True,
    'provisioning_status': 'ACTIVE'}

RET_MEMBER_2 = {
    'id': 'sample_member_id_2',
    'address': '10.0.0.98',
    'protocol_port': 82,
    'weight': 13,
    'subnet_id': '10.0.0.1/24',
    'admin_state_up': True,
    'provisioning_status': 'ACTIVE'}

RET_POOL = {
    'id': 'sample_pool_id_1',
    'protocol': 'http',
    'lb_algorithm': 'roundrobin',
    'members': [RET_MEMBER_1, RET_MEMBER_2],
    'health_monitor': RET_MONITOR,
    'session_persistence': RET_PERSISTENCE,
    'admin_state_up': True,
    'provisioning_status': 'ACTIVE'}

RET_DEF_TLS_CONT = {'id': 'cont_id_1', 'allencompassingpem': 'imapem'}
RET_SNI_CONT_1 = {'id': 'cont_id_2', 'allencompassingpem': 'imapem2'}
RET_SNI_CONT_2 = {'id': 'cont_id_3', 'allencompassingpem': 'imapem3'}

RET_LISTENER = {
    'id': 'sample_listener_id_1',
    'protocol_port': '80',
    'protocol': 'HTTP',
    'protocol_mode': 'http',
    'default_pool': RET_POOL,
    'connection_limit': 98}

RET_LISTENER_TLS = {
    'id': 'sample_listener_id_1',
    'protocol_port': '443',
    'protocol_mode': 'http',
    'protocol': 'TERMINATED_HTTPS',
    'default_pool': RET_POOL,
    'connection_limit': 98,
    'default_tls_container_id': 'cont_id_1',
    'default_tls_path': '/v2/sample_loadbalancer_id_1/cont_id_1.pem',
    'default_tls_container': RET_DEF_TLS_CONT}

RET_LISTENER_TLS_SNI = {
    'id': 'sample_listener_id_1',
    'protocol_port': '443',
    'protocol_mode': 'http',
    'protocol': 'TERMINATED_HTTPS',
    'default_pool': RET_POOL,
    'connection_limit': 98,
    'default_tls_container_id': 'cont_id_1',
    'default_tls_path': '/v2/sample_loadbalancer_id_1/cont_id_1.pem',
    'default_tls_container': RET_DEF_TLS_CONT,
    'crt_dir': '/v2/sample_loadbalancer_id_1',
    'sni_container_ids': ['cont_id_2', 'cont_id_3'],
    'sni_containers': [RET_SNI_CONT_1, RET_SNI_CONT_2]}

RET_LB = {
    'id': 'sample_loadbalancer_id_1',
    'vip_address': '10.0.0.2',
    'listeners': [RET_LISTENER],
    'connection_limit': RET_LISTENER['connection_limit'],
    'pools': [RET_POOL]}

RET_LB_TLS = {
    'id': 'sample_loadbalancer_id_1',
    'vip_address': '10.0.0.2',
    'listeners': [RET_LISTENER_TLS],
    'connection_limit': RET_LISTENER_TLS['connection_limit'],
    'pools': [RET_POOL]}

RET_LB_TLS_SNI = {
    'id': 'sample_loadbalancer_id_1',
    'vip_address': '10.0.0.2',
    'listeners': [RET_LISTENER_TLS_SNI],
    'connection_limit': RET_LISTENER_TLS_SNI['connection_limit'],
    'pools': [RET_POOL]}


def sample_loadbalancer_tuple(proto=None, monitor=True, persistence=True,
                              persistence_type=None, tls=False, sni=False):
    proto = 'HTTP' if proto is None else proto
    in_lb = collections.namedtuple(
        'loadbalancer', 'id, vip_address, protocol, vip_port, '
                        'listeners, pools')
    return in_lb(
        id='sample_loadbalancer_id_1',
        vip_address='10.0.0.2',
        protocol=proto,
        vip_port=sample_vip_port_tuple(),
        listeners=[sample_listener_tuple(proto=proto, monitor=monitor,
                                         persistence=persistence,
                                         persistence_type=persistence_type,
                                         tls=tls,
                                         sni=sni)],
        pools=[sample_pool_tuple(proto=proto, monitor=monitor,
                                 persistence=persistence,
                                 persistence_type=persistence_type)]
    )


def sample_vip_port_tuple():
    vip_port = collections.namedtuple('vip_port', 'fixed_ips')
    ip_address = collections.namedtuple('ip_address', 'ip_address')
    in_address = ip_address(ip_address='10.0.0.2')
    return vip_port(fixed_ips=[in_address])


def sample_listener_tuple(proto=None, monitor=True, persistence=True,
                          persistence_type=None, tls=False, sni=False,
                          connection_limit=98):
    proto = 'HTTP' if proto is None else proto
    port = '443' if proto is 'HTTPS' or proto is 'TERMINATED_HTTPS' else '80'
    in_listener = collections.namedtuple(
        'listener', 'id, tenant_id, protocol_port, protocol, default_pool, '
        'connection_limit, admin_state_up, default_tls_container_id, '
                    'sni_container_ids, default_tls_container, '
                    'sni_containers, loadbalancer_id')
    return in_listener(
        id='sample_listener_id_1',
        tenant_id='sample_tenant_id',
        protocol_port=port,
        protocol=proto,
        default_pool=sample_pool_tuple(
            proto=proto, monitor=monitor, persistence=persistence,
            persistence_type=persistence_type),
        connection_limit=connection_limit,
        admin_state_up=True,
        default_tls_container_id='cont_id_1' if tls else '',
        sni_container_ids=['cont_id_2', 'cont_id_3'] if sni else [],
        default_tls_container=sample_tls_container_tuple(
            id='cont_id_1', certificate='--imapem1--\n',
            private_key='--imakey1--\n', intermediates=[
                '--imainter1--\n', '--imainter1too--\n'],
            primary_cn='fakeCNM'
        ) if tls else '',
        sni_containers=[
            sample_tls_sni_container_tuple(
                tls_container_id='cont_id_2',
                tls_container=sample_tls_container_tuple(
                    id='cont_id_2', certificate='--imapem2--\n',
                    private_key='--imakey2--\n', intermediates=[
                        '--imainter2--\n', '--imainter2too--\n'],
                    primary_cn='fakeCN')),
            sample_tls_sni_container_tuple(
                tls_container_id='cont_id_3',
                tls_container=sample_tls_container_tuple(
                    id='cont_id_3', certificate='--imapem3--\n',
                    private_key='--imakey3--\n', intermediates=[
                        '--imainter3--\n', '--imainter3too--\n'],
                    primary_cn='fakeCN2'))]
        if sni else [],
        loadbalancer_id='sample_loadbalancer_id_1'
    )


def sample_tls_sni_container_tuple(tls_container=None, tls_container_id=None):
    sc = collections.namedtuple('sni_container', 'tls_container,'
                                                 'tls_container_id')
    return sc(tls_container=tls_container, tls_container_id=tls_container_id)


def sample_tls_container_tuple(id='cont_id_1', certificate=None,
                               private_key=None, intermediates=None,
                               primary_cn=None):
    intermediates = intermediates or []
    sc = collections.namedtuple(
        'tls_cert',
        'id, certificate, private_key, intermediates, primary_cn')
    return sc(id=id, certificate=certificate, private_key=private_key,
              intermediates=intermediates or [], primary_cn=primary_cn)


def sample_pool_tuple(proto=None, monitor=True, persistence=True,
                      persistence_type=None, hm_admin_state=True):
    proto = 'HTTP' if proto is None else proto
    in_pool = collections.namedtuple(
        'pool', 'id, protocol, lb_algorithm, members, healthmonitor,'
                'session_persistence, admin_state_up, provisioning_status')
    mon = (sample_health_monitor_tuple(proto=proto, admin_state=hm_admin_state)
           if monitor is True else None)
    persis = sample_session_persistence_tuple(
        persistence_type=persistence_type) if persistence is True else None
    return in_pool(
        id='sample_pool_id_1',
        protocol=proto,
        lb_algorithm='ROUND_ROBIN',
        members=[sample_member_tuple('sample_member_id_1', '10.0.0.99'),
                 sample_member_tuple('sample_member_id_2', '10.0.0.98')],
        healthmonitor=mon,
        session_persistence=persis,
        admin_state_up=True,
        provisioning_status='ACTIVE')


def sample_member_tuple(id, ip, admin_state_up=True, status='ACTIVE'):
    in_member = collections.namedtuple('member',
                                       'id, address, protocol_port, '
                                       'weight, subnet_id, '
                                       'admin_state_up, provisioning_status')
    return in_member(
        id=id,
        address=ip,
        protocol_port=82,
        weight=13,
        subnet_id='10.0.0.1/24',
        admin_state_up=admin_state_up,
        provisioning_status=status)


def sample_session_persistence_tuple(persistence_type=None):
    spersistence = collections.namedtuple('SessionPersistence',
                                          'type, cookie_name')
    pt = 'HTTP_COOKIE' if persistence_type is None else persistence_type
    return spersistence(type=pt,
                        cookie_name=pt)


def sample_health_monitor_tuple(proto='HTTP', admin_state=True):
    proto = 'HTTP' if proto is 'TERMINATED_HTTPS' else proto
    monitor = collections.namedtuple(
        'monitor', 'id, type, delay, timeout, max_retries, http_method, '
                   'url_path, expected_codes, admin_state_up')

    return monitor(id='sample_monitor_id_1', type=proto, delay=30,
                   timeout=31, max_retries=3, http_method='GET',
                   url_path='/index.html', expected_codes='500, 405, 404',
                   admin_state_up=admin_state)


def sample_base_expected_config(backend, frontend=None,
                                fe_proto=constants.PROTOCOL_HTTP):
    if frontend is None:
        tcp_frontend = ("frontend sample_listener_id_1\n"
                        "    option tcplog\n"
                        "    maxconn 98\n"
                        "    bind 10.0.0.2:80\n"
                        "    mode tcp\n"
                        "    default_backend sample_pool_id_1\n\n")
        http_frontend = ("frontend sample_listener_id_1\n"
                         "    option httplog\n"
                         "    maxconn 98\n"
                         "    option forwardfor\n"
                         "    bind 10.0.0.2:80\n"
                         "    mode http\n"
                         "    default_backend sample_pool_id_1\n\n")
        https_frontend = ("frontend sample_listener_id_1\n"
                          "    option tcplog\n"
                          "    maxconn 98\n"
                          "    bind 10.0.0.2:443\n"
                          "    mode tcp\n"
                          "    default_backend sample_pool_id_1\n\n")
        https_tls_frontend = ("frontend sample_listener_id_1\n"
                              "    option httplog\n"
                              "    redirect scheme https if !{ ssl_fc }\n"
                              "    maxconn 98\n"
                              "    option forwardfor\n"
                              "    bind 10.0.0.2:443\n"
                              "    mode http\n"
                              "    default_backend sample_pool_id_1\n\n")
        fe_mapper = {
            constants.PROTOCOL_TCP: tcp_frontend,
            constants.PROTOCOL_HTTP: http_frontend,
            constants.PROTOCOL_HTTPS: https_frontend,
            constants.PROTOCOL_TERMINATED_HTTPS: https_tls_frontend
        }
        frontend = fe_mapper[fe_proto]
    return ("# Configuration for sample_loadbalancer_id_1\n"
            "global\n"
            "    daemon\n"
            "    user nobody\n"
            "    group nogroup\n"
            "    log /dev/log local0 debug alert\n"
            "    log /dev/log local1 notice alert\n"
            "    maxconn 98\n"
            "    stats socket /sock_path mode 0666 level user\n\n"
            "defaults\n"
            "    log global\n"
            "    retries 3\n"
            "    option redispatch\n"
            "    timeout connect 5000\n"
            "    timeout client 50000\n"
            "    timeout server 50000\n\n" + frontend + backend)



#
# Copyright 2014-2015 Rackspace.  All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import netaddr
from neutron.common import ipv6_utils
from neutron.db import common_db_mixin as base_db
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as pg_const
from neutron_lib.plugins import directory
from oslo_db import exception
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
from sqlalchemy import orm
from sqlalchemy.orm import exc
from sqlalchemy.orm import lazyload
from sqlalchemy.orm import subqueryload

from neutron_lbaas._i18n import _
from neutron_lbaas import agent_scheduler
from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.extensions import l7
from neutron_lbaas.extensions import lb_network_vip
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.extensions import sharedpools
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import data_models


LOG = logging.getLogger(__name__)


class LoadBalancerPluginDbv2(base_db.CommonDbMixin,
                             agent_scheduler.LbaasAgentSchedulerDbMixin):
    """Wraps loadbalancer with SQLAlchemy models.

    A class that wraps the implementation of the Neutron loadbalancer
    plugin database access interface using SQLAlchemy models.
    """

    @property
    def _core_plugin(self):
        return directory.get_plugin()

    def _get_resource(self, context, model, id, for_update=False):
        resource = None
        try:
            if for_update:
                # To lock the instance for update, return a single
                # instance, instead of an instance with LEFT OUTER
                # JOINs that do not work in PostgreSQL
                query = self._model_query(context, model).options(
                    lazyload('*')
                ).filter(
                    model.id == id).with_lockmode('update')
                resource = query.one()
            else:
                resource = self._get_by_id(context, model, id)
        except exc.NoResultFound:
            with excutils.save_and_reraise_exception(reraise=False) as ctx:
                if issubclass(model, (models.LoadBalancer, models.Listener,
                                      models.L7Policy, models.L7Rule,
                                      models.PoolV2, models.MemberV2,
                                      models.HealthMonitorV2,
                                      models.LoadBalancerStatistics,
                                      models.SessionPersistenceV2)):
                    raise loadbalancerv2.EntityNotFound(name=model.NAME, id=id)
                ctx.reraise = True
        return resource

    def _resource_exists(self, context, model, id):
        try:
            self._get_by_id(context, model, id)
        except exc.NoResultFound:
            return False
        return True

    def _get_resources(self, context, model, filters=None, options=None):
        query = self._get_collection_query(context, model,
                                           filters=filters)
        if options:
            query = query.options(options)
        return [model_instance for model_instance in query]

    def _create_port_choose_fixed_ip(self, fixed_ips):
        # Neutron will try to allocate IPv4, IPv6, and IPv6 EUI-64 addresses.
        # We're most interested in the IPv4 address. An IPv4 vip can be
        # routable from IPv6. Creating a port by network can be used to manage
        # the dwindling, fragmented IPv4 address space. IPv6 has enough
        # addresses that a single subnet can always be created that's big
        # enough to allocate all vips.
        for fixed_ip in fixed_ips:
            ip_address = fixed_ip['ip_address']
            ip = netaddr.IPAddress(ip_address)
            if ip.version == 4:
                return fixed_ip
        # An EUI-64 address isn't useful as a vip
        for fixed_ip in fixed_ips:
            ip_address = fixed_ip['ip_address']
            ip = netaddr.IPAddress(ip_address)
            if ip.version == 6 and not ipv6_utils.is_eui64_address(ip_address):
                return fixed_ip
        for fixed_ip in fixed_ips:
            return fixed_ip

    def _create_port_for_load_balancer(self, context, lb_db, ip_address,
                                       network_id=None):
        if lb_db.vip_subnet_id:
            assign_subnet = False
            # resolve subnet and create port
            subnet = self._core_plugin.get_subnet(context, lb_db.vip_subnet_id)
            network_id = subnet['network_id']
            fixed_ip = {'subnet_id': subnet['id']}
            if ip_address and ip_address != n_const.ATTR_NOT_SPECIFIED:
                fixed_ip['ip_address'] = ip_address
            fixed_ips = [fixed_ip]
        elif network_id and network_id != n_const.ATTR_NOT_SPECIFIED:
            assign_subnet = True
            fixed_ips = n_const.ATTR_NOT_SPECIFIED
        else:
            attrs = _("vip_subnet_id or vip_network_id")
            raise loadbalancerv2.RequiredAttributeNotSpecified(attr_name=attrs)

        port_data = {
            'tenant_id': lb_db.tenant_id,
            'name': 'loadbalancer-' + lb_db.id,
            'network_id': network_id,
            'mac_address': n_const.ATTR_NOT_SPECIFIED,
            'admin_state_up': False,
            'device_id': lb_db.id,
            'device_owner': n_const.DEVICE_OWNER_LOADBALANCERV2,
            'fixed_ips': fixed_ips
        }

        port = self._core_plugin.create_port(context, {'port': port_data})
        lb_db.vip_port_id = port['id']

        if assign_subnet:
            fixed_ip = self._create_port_choose_fixed_ip(port['fixed_ips'])
            lb_db.vip_address = fixed_ip['ip_address']
            lb_db.vip_subnet_id = fixed_ip['subnet_id']
        else:
            for fixed_ip in port['fixed_ips']:
                if fixed_ip['subnet_id'] == lb_db.vip_subnet_id:
                    lb_db.vip_address = fixed_ip['ip_address']
                    break

    def _create_loadbalancer_stats(self, context, loadbalancer_id, data=None):
        # This is internal method to add load balancer statistics.  It won't
        # be exposed to API
        data = data or {}
        stats_db = models.LoadBalancerStatistics(
            loadbalancer_id=loadbalancer_id,
            bytes_in=data.get(lb_const.STATS_IN_BYTES, 0),
            bytes_out=data.get(lb_const.STATS_OUT_BYTES, 0),
            active_connections=data.get(lb_const.STATS_ACTIVE_CONNECTIONS, 0),
            total_connections=data.get(lb_const.STATS_TOTAL_CONNECTIONS, 0)
        )
        return stats_db

    def _delete_loadbalancer_stats(self, context, loadbalancer_id):
        # This is internal method to delete pool statistics. It won't
        # be exposed to API
        with context.session.begin(subtransactions=True):
            stats_qry = context.session.query(models.LoadBalancerStatistics)
            try:
                stats = stats_qry.filter_by(
                    loadbalancer_id=loadbalancer_id).one()
            except exc.NoResultFound:
                raise loadbalancerv2.EntityNotFound(
                    name=models.LoadBalancerStatistics.NAME,
                    id=loadbalancer_id)
            context.session.delete(stats)

    def _load_id(self, context, model_dict):
        model_dict['id'] = uuidutils.generate_uuid()

    def check_subnet_exists(self, context, subnet_id):
        try:
            self._core_plugin.get_subnet(context, subnet_id)
        except n_exc.SubnetNotFound:
            raise loadbalancerv2.EntityNotFound(name="Subnet", id=subnet_id)

    def _validate_and_return_vip_net(self, ctxt, lb):
        network_id = lb.pop('vip_network_id', None)

        if network_id != n_const.ATTR_NOT_SPECIFIED and network_id:
            subnets = self._core_plugin.get_subnets_by_network(ctxt,
                                                               network_id)
            if not subnets:
                raise lb_network_vip.VipNetworkInvalid(network=network_id)
            return network_id
        return

    def assert_modification_allowed(self, obj):
        status = getattr(obj, 'provisioning_status', None)
        if status in [n_const.PENDING_DELETE, n_const.PENDING_UPDATE,
                      n_const.PENDING_CREATE]:
            id = getattr(obj, 'id', None)
            raise loadbalancerv2.StateInvalid(id=id, state=status)

    def test_and_set_status(self, context, model, id, status):
        with context.session.begin(subtransactions=True):
            db_lb_child = None
            if model == models.LoadBalancer:
                db_lb = self._get_resource(context, model, id, for_update=True)
            else:
                db_lb_child = self._get_resource(context, model, id)
                db_lb = self._get_resource(context, models.LoadBalancer,
                                           db_lb_child.root_loadbalancer.id)
            # This method will raise an exception if modification is not
            # allowed.
            self.assert_modification_allowed(db_lb)

            # if the model passed in is not a load balancer then we will
            # set its root load balancer's provisioning status to
            # PENDING_UPDATE and the model's status to the status passed in
            # Otherwise we are just setting the load balancer's provisioning
            # status to the status passed in
            if db_lb_child:
                db_lb.provisioning_status = n_const.PENDING_UPDATE
                db_lb_child.provisioning_status = status
            else:
                db_lb.provisioning_status = status

    def update_loadbalancer_provisioning_status(self, context, lb_id,
                                                status=n_const.ACTIVE):
        self.update_status(context, models.LoadBalancer, lb_id,
                           provisioning_status=status)

    def update_status(self, context, model, id, provisioning_status=None,
                      operating_status=None):
        with context.session.begin(subtransactions=True):
            if issubclass(model, models.LoadBalancer):
                try:
                    model_db = (self._model_query(context, model).
                                filter(model.id == id).
                                options(orm.noload('vip_port')).
                                one())
                except exc.NoResultFound:
                    raise loadbalancerv2.EntityNotFound(
                        name=models.LoadBalancer.NAME, id=id)
            else:
                model_db = self._get_resource(context, model, id)
            if provisioning_status and (model_db.provisioning_status !=
                                        provisioning_status):
                model_db.provisioning_status = provisioning_status
                LOG.debug("Provisioning status of %s (id=%s) updated to: %s",
                          model_db.NAME, model_db.id, provisioning_status)
            if (operating_status and hasattr(model_db, 'operating_status') and
                    model_db.operating_status != operating_status):
                model_db.operating_status = operating_status
                LOG.debug("Operating status of %s (id=%s) updated to: %s",
                          model_db.NAME, model_db.id, operating_status)

    def create_loadbalancer_graph(self, context, loadbalancer,
                                  allocate_vip=True):
        l7policies_ids = []
        with context.session.begin(subtransactions=True):
            listeners = loadbalancer.pop('listeners', [])
            lb_db = self.create_loadbalancer(context, loadbalancer,
                                             allocate_vip=allocate_vip)
            for listener in listeners:
                listener['loadbalancer_id'] = lb_db.id
                default_pool = listener.pop('default_pool', None)
                if (default_pool and
                        default_pool != n_const.ATTR_NOT_SPECIFIED):
                    default_pool['loadbalancer_id'] = lb_db.id
                    hm = default_pool.pop('healthmonitor', None)
                    if hm and hm != n_const.ATTR_NOT_SPECIFIED:
                        hm_db = self.create_healthmonitor(context, hm)
                        default_pool['healthmonitor_id'] = hm_db.id
                    members = default_pool.pop('members', [])
                    pool_db = self.create_pool(context, default_pool)
                    listener['default_pool_id'] = pool_db.id
                    for member in members:
                        member['pool_id'] = pool_db.id
                        self.create_pool_member(context, member, pool_db.id)
                l7policies = listener.pop('l7policies', None)
                listener_db = self.create_listener(context, listener)
                if (l7policies and l7policies !=
                        n_const.ATTR_NOT_SPECIFIED):
                    for l7policy in l7policies:
                        l7policy['listener_id'] = listener_db.id
                        redirect_pool = l7policy.pop('redirect_pool', None)
                        l7rules = l7policy.pop('rules', [])
                        if (redirect_pool and redirect_pool !=
                                n_const.ATTR_NOT_SPECIFIED):
                            redirect_pool['loadbalancer_id'] = lb_db.id
                            rhm = redirect_pool.pop('healthmonitor', None)
                            rmembers = redirect_pool.pop('members', [])
                            if rhm and rhm != n_const.ATTR_NOT_SPECIFIED:
                                rhm_db = self.create_healthmonitor(context,
                                                                   rhm)
                                redirect_pool['healthmonitor_id'] = rhm_db.id
                            rpool_db = self.create_pool(context, redirect_pool)
                            l7policy['redirect_pool_id'] = rpool_db.id
                            for rmember in rmembers:
                                rmember['pool_id'] = rpool_db.id
                                self.create_pool_member(context, rmember,
                                                        rpool_db.id)
                        l7policy_db = self.create_l7policy(context, l7policy)
                        l7policies_ids.append(l7policy_db.id)
                        if (l7rules and l7rules !=
                                n_const.ATTR_NOT_SPECIFIED):
                            for l7rule in l7rules:
                                self.create_l7policy_rule(
                                    context, l7rule, l7policy_db.id)
        # SQL Alchemy cache issue where l7rules won't show up as intended.
        for l7policy_id in l7policies_ids:
            l7policy_db = self._get_resource(context, models.L7Policy,
                                           l7policy_id)
            context.session.expire(l7policy_db)
        return self.get_loadbalancer(context, lb_db.id)

    def create_loadbalancer(self, context, loadbalancer, allocate_vip=True):
        self._load_id(context, loadbalancer)
        vip_network_id = self._validate_and_return_vip_net(context,
                                                           loadbalancer)
        vip_subnet_id = loadbalancer.pop('vip_subnet_id', None)
        vip_address = loadbalancer.pop('vip_address')
        if vip_subnet_id and vip_subnet_id != n_const.ATTR_NOT_SPECIFIED:
            loadbalancer['vip_subnet_id'] = vip_subnet_id
        loadbalancer['provisioning_status'] = n_const.PENDING_CREATE
        loadbalancer['operating_status'] = lb_const.OFFLINE
        lb_db = models.LoadBalancer(**loadbalancer)

        # create port outside of lb create transaction since it can sometimes
        # cause lock wait timeouts
        if allocate_vip:
            LOG.debug("Plugin will allocate the vip as a neutron port.")
            self._create_port_for_load_balancer(context, lb_db,
                                                vip_address, vip_network_id)

        with context.session.begin(subtransactions=True):
            context.session.add(lb_db)
            context.session.flush()
            lb_db.stats = self._create_loadbalancer_stats(
                context, lb_db.id)
            context.session.add(lb_db)
            context.session.flush()
        return data_models.LoadBalancer.from_sqlalchemy_model(lb_db)

    def update_loadbalancer(self, context, id, loadbalancer):
        with context.session.begin(subtransactions=True):
            lb_db = self._get_resource(context, models.LoadBalancer, id)
            lb_db.update(loadbalancer)
        return data_models.LoadBalancer.from_sqlalchemy_model(lb_db)

    def delete_loadbalancer(self, context, id, delete_vip_port=True):
        with context.session.begin(subtransactions=True):
            lb_db = self._get_resource(context, models.LoadBalancer, id)
            context.session.delete(lb_db)
        if delete_vip_port and lb_db.vip_port:
            self._delete_vip_port(context, lb_db.vip_port_id)

    @db_api.retry_db_errors
    def _delete_vip_port(self, context, vip_port_id):
        self._core_plugin.delete_port(context, vip_port_id)

    def prevent_lbaasv2_port_deletion(self, context, port_id):
        try:
            port_db = self._core_plugin._get_port(context, port_id)
        except n_exc.PortNotFound:
            return
        if port_db['device_owner'] == n_const.DEVICE_OWNER_LOADBALANCERV2:
            filters = {'vip_port_id': [port_id]}
            if len(self.get_loadbalancer_ids(context, filters=filters)) > 0:
                reason = _('has device owner %s') % port_db['device_owner']
                raise n_exc.ServicePortInUse(port_id=port_db['id'],
                                             reason=reason)

    def subscribe(self):
        registry.subscribe(
            _prevent_lbaasv2_port_delete_callback, resources.PORT,
            events.BEFORE_DELETE)

    def get_loadbalancer_ids(self, context, filters=None):
        lb_dbs = self._get_resources(context, models.LoadBalancer,
                                     filters=filters)
        return [lb_db.id
                for lb_db in lb_dbs]

    def get_loadbalancers(self, context, filters=None):
        lb_dbs = self._get_resources(context, models.LoadBalancer,
                                     filters=filters)
        return [data_models.LoadBalancer.from_sqlalchemy_model(lb_db)
                for lb_db in lb_dbs]

    def get_loadbalancers_as_api_dict(self, context, filters=None):
        options = (
            subqueryload(models.LoadBalancer.listeners),
            subqueryload(models.LoadBalancer.pools),
            subqueryload(models.LoadBalancer.provider)
        )
        lb_dbs = self._get_resources(context, models.LoadBalancer,
                                     filters=filters, options=options)
        return [lb_db.to_api_dict
                for lb_db in lb_dbs]

    def get_provider_names_used_in_loadbalancers(self, context):
        lb_dbs = self._get_resources(context, models.LoadBalancer)
        return [lb_db.provider.provider_name for lb_db in lb_dbs]

    def get_loadbalancer(self, context, id):
        lb_db = self._get_resource(context, models.LoadBalancer, id)
        return data_models.LoadBalancer.from_sqlalchemy_model(lb_db)

    def _validate_listener_data(self, context, listener):
        pool_id = listener.get('default_pool_id')
        lb_id = listener.get('loadbalancer_id')
        if lb_id:
            if not self._resource_exists(context, models.LoadBalancer,
                                         lb_id):
                raise loadbalancerv2.EntityNotFound(
                    name=models.LoadBalancer.NAME, id=lb_id)
        if pool_id:
            if not self._resource_exists(context, models.PoolV2, pool_id):
                raise loadbalancerv2.EntityNotFound(
                    name=models.PoolV2.NAME, id=pool_id)
            pool = self._get_resource(context, models.PoolV2, pool_id)
            if ((pool.protocol, listener.get('protocol'))
                not in lb_const.LISTENER_POOL_COMPATIBLE_PROTOCOLS):
                raise loadbalancerv2.ListenerPoolProtocolMismatch(
                    listener_proto=listener['protocol'],
                    pool_proto=pool.protocol)
        if lb_id and pool_id:
            pool = self._get_resource(context, models.PoolV2, pool_id)
            if pool.loadbalancer_id != lb_id:
                raise sharedpools.ListenerPoolLoadbalancerMismatch(
                    pool_id=pool_id,
                    lb_id=pool.loadbalancer_id)

    def _validate_l7policy_data(self, context, l7policy):
        if l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL:
            if not l7policy['redirect_pool_id']:
                raise l7.L7PolicyRedirectPoolIdMissing()
            if not self._resource_exists(
                context, models.PoolV2, l7policy['redirect_pool_id']):
                raise loadbalancerv2.EntityNotFound(
                    name=models.PoolV2.NAME, id=l7policy['redirect_pool_id'])

            pool = self._get_resource(
                context, models.PoolV2, l7policy['redirect_pool_id'])

            listener = self._get_resource(
                context, models.Listener, l7policy['listener_id'])

            if pool.loadbalancer_id != listener.loadbalancer_id:
                raise sharedpools.ListenerAndPoolMustBeOnSameLoadbalancer()

        if (l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL and
            'redirect_url' not in l7policy):
            raise l7.L7PolicyRedirectUrlMissing()

    def _validate_l7rule_data(self, context, rule):
        def _validate_regex(regex):
            try:
                re.compile(regex)
            except Exception as e:
                raise l7.L7RuleInvalidRegex(e=str(e))

        def _validate_key(key):
            p = re.compile(lb_const.HTTP_HEADER_COOKIE_NAME_REGEX)
            if not p.match(key):
                raise l7.L7RuleInvalidKey()

        def _validate_cookie_value(value):
            p = re.compile(lb_const.HTTP_COOKIE_VALUE_REGEX)
            if not p.match(value):
                raise l7.L7RuleInvalidCookieValue()

        def _validate_non_cookie_value(value):
            p = re.compile(lb_const.HTTP_HEADER_VALUE_REGEX)
            q = re.compile(lb_const.HTTP_QUOTED_HEADER_VALUE_REGEX)
            if not p.match(value) and not q.match(value):
                raise l7.L7RuleInvalidHeaderValue()

        if rule['compare_type'] == lb_const.L7_RULE_COMPARE_TYPE_REGEX:
            _validate_regex(rule['value'])

        if rule['type'] in [lb_const.L7_RULE_TYPE_HEADER,
                            lb_const.L7_RULE_TYPE_COOKIE]:
            if ('key' not in rule or not rule['key']):
                raise l7.L7RuleKeyMissing()
            _validate_key(rule['key'])

        if rule['compare_type'] != lb_const.L7_RULE_COMPARE_TYPE_REGEX:
            if rule['type'] == lb_const.L7_RULE_TYPE_COOKIE:
                _validate_cookie_value(rule['value'])
            else:
                if rule['type'] in [lb_const.L7_RULE_TYPE_HEADER,
                                  lb_const.L7_RULE_TYPE_HOST_NAME,
                                  lb_const.L7_RULE_TYPE_PATH]:
                    _validate_non_cookie_value(rule['value'])
                elif (rule['compare_type'] ==
                      lb_const.L7_RULE_COMPARE_TYPE_EQUAL_TO):
                    _validate_non_cookie_value(rule['value'])
                else:
                    raise l7.L7RuleUnsupportedCompareType(type=rule['type'])

    def _convert_api_to_db(self, listener):
        # NOTE(blogan): Converting the values for db models for now to
        # limit the scope of this change
        if 'default_tls_container_ref' in listener:
            tls_cref = listener.get('default_tls_container_ref')
            del listener['default_tls_container_ref']
            listener['default_tls_container_id'] = tls_cref
        if 'sni_container_refs' in listener:
            sni_crefs = listener.get('sni_container_refs')
            del listener['sni_container_refs']
            listener['sni_container_ids'] = sni_crefs

    def create_listener(self, context, listener):
        self._convert_api_to_db(listener)
        try:
            with context.session.begin(subtransactions=True):
                self._load_id(context, listener)
                listener['provisioning_status'] = n_const.PENDING_CREATE
                listener['operating_status'] = lb_const.OFFLINE
                # Check for unspecified loadbalancer_id and listener_id and
                # set to None
                for id in ['loadbalancer_id', 'default_pool_id']:
                    if listener.get(id) == n_const.ATTR_NOT_SPECIFIED:
                        listener[id] = None

                self._validate_listener_data(context, listener)
                sni_container_ids = []
                if 'sni_container_ids' in listener:
                    sni_container_ids = listener.pop('sni_container_ids')
                try:
                    listener_db_entry = models.Listener(**listener)
                except Exception as exc:
                    raise exc
                for container_id in sni_container_ids:
                    sni = models.SNI(listener_id=listener_db_entry.id,
                                     tls_container_id=container_id)
                    listener_db_entry.sni_containers.append(sni)
                context.session.add(listener_db_entry)
        except exception.DBDuplicateEntry:
            raise loadbalancerv2.LoadBalancerListenerProtocolPortExists(
                lb_id=listener['loadbalancer_id'],
                protocol_port=listener['protocol_port'])
        context.session.refresh(listener_db_entry.loadbalancer)
        return data_models.Listener.from_sqlalchemy_model(listener_db_entry)

    def update_listener(self, context, id, listener,
                        tls_containers_changed=False):
        self._convert_api_to_db(listener)
        with context.session.begin(subtransactions=True):
            listener_db = self._get_resource(context, models.Listener, id)

            if not listener.get('protocol'):
                # User did not intend to change the protocol so we will just
                # use the same protocol already stored so the validation knows
                listener['protocol'] = listener_db.protocol
            self._validate_listener_data(context, listener)

            if tls_containers_changed:
                listener_db.sni_containers = []
                for container_id in listener['sni_container_ids']:
                    sni = models.SNI(listener_id=id,
                                     tls_container_id=container_id)
                    listener_db.sni_containers.append(sni)

            listener_db.update(listener)

        context.session.refresh(listener_db)
        return data_models.Listener.from_sqlalchemy_model(listener_db)

    def delete_listener(self, context, id):
        listener_db_entry = self._get_resource(context, models.Listener, id)
        with context.session.begin(subtransactions=True):
            context.session.delete(listener_db_entry)

    def get_listeners(self, context, filters=None):
        listener_dbs = self._get_resources(context, models.Listener,
                                           filters=filters)
        return [data_models.Listener.from_sqlalchemy_model(listener_db)
                for listener_db in listener_dbs]

    def get_listeners_as_api_dict(self, context, filters=None):
        options = (
            subqueryload(models.Listener.sni_containers),
            subqueryload(models.Listener.loadbalancer),
            subqueryload(models.Listener.l7_policies)
        )
        listener_dbs = self._get_resources(context, models.Listener,
                                           filters=filters, options=options)
        return [listener_db.to_api_dict
                for listener_db in listener_dbs]

    def get_listener(self, context, id):
        listener_db = self._get_resource(context, models.Listener, id)
        return data_models.Listener.from_sqlalchemy_model(listener_db)

    def _create_session_persistence_db(self, session_info, pool_id):
        session_info['pool_id'] = pool_id
        return models.SessionPersistenceV2(**session_info)

    def _update_pool_session_persistence(self, context, pool_id, info):
        # removing these keys as it is possible that they are passed in and
        # their existence will cause issues bc they are not acceptable as
        # dictionary values
        info.pop('pool', None)
        info.pop('pool_id', None)
        pool = self._get_resource(context, models.PoolV2, pool_id)
        with context.session.begin(subtransactions=True):
            # Update sessionPersistence table
            sess_qry = context.session.query(models.SessionPersistenceV2)
            sesspersist_db = sess_qry.filter_by(pool_id=pool_id).first()

            # Insert a None cookie_info if it is not present to overwrite an
            # existing value in the database.
            if 'cookie_name' not in info:
                info['cookie_name'] = None

            if sesspersist_db:
                sesspersist_db.update(info)
            else:
                info['pool_id'] = pool_id
                sesspersist_db = models.SessionPersistenceV2(**info)
                context.session.add(sesspersist_db)
                # Update pool table
                pool.session_persistence = sesspersist_db
            context.session.add(pool)

    def _delete_session_persistence(self, context, pool_id):
        with context.session.begin(subtransactions=True):
            sess_qry = context.session.query(models.SessionPersistenceV2)
            sess_qry.filter_by(pool_id=pool_id).delete()

    def create_pool(self, context, pool):
        with context.session.begin(subtransactions=True):
            self._load_id(context, pool)
            pool['provisioning_status'] = n_const.PENDING_CREATE
            pool['operating_status'] = lb_const.OFFLINE

            session_info = pool.pop('session_persistence', None)
            pool_db = models.PoolV2(**pool)

            if session_info:
                s_p = self._create_session_persistence_db(session_info,
                                                          pool_db.id)
                pool_db.session_persistence = s_p

            context.session.add(pool_db)
        context.session.refresh(pool_db.loadbalancer)
        return self.get_pool(context, pool_db.id)

    def update_pool(self, context, id, pool):
        with context.session.begin(subtransactions=True):
            pool_db = self._get_resource(context, models.PoolV2, id)
            hm_id = pool.get('healthmonitor_id')
            if hm_id:
                if not self._resource_exists(context, models.HealthMonitorV2,
                                             hm_id):
                    raise loadbalancerv2.EntityNotFound(
                        name=models.HealthMonitorV2.NAME,
                        id=hm_id)
                filters = {'healthmonitor_id': [hm_id]}
                hmpools = self._get_resources(context,
                                              models.PoolV2,
                                              filters=filters)
                if hmpools:
                    raise loadbalancerv2.EntityInUse(
                        entity_using=models.PoolV2.NAME,
                        id=hmpools[0].id,
                        entity_in_use=models.HealthMonitorV2.NAME)

            # Only update or delete session persistence if it was part
            # of the API request.
            if 'session_persistence' in pool.keys():
                sp = pool.pop('session_persistence')
                if sp is None or sp == {}:
                    self._delete_session_persistence(context, id)
                else:
                    self._update_pool_session_persistence(context, id, sp)

            # sqlalchemy cries if listeners is defined.
            listeners = pool.get('listeners')
            if listeners:
                del pool['listeners']
            pool_db.update(pool)
        context.session.refresh(pool_db)
        return data_models.Pool.from_sqlalchemy_model(pool_db)

    def delete_pool(self, context, id):
        with context.session.begin(subtransactions=True):
            pool_db = self._get_resource(context, models.PoolV2, id)
            for l in pool_db.listeners:
                self.update_listener(context, l.id,
                                     {'default_pool_id': None})
            for l in pool_db.loadbalancer.listeners:
                for p in l.l7_policies:
                    if (p.action ==
                        lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL and
                        p.redirect_pool_id == id):
                        self.update_l7policy(
                            context, p.id,
                            {'redirect_pool_id': None,
                             'action': lb_const.L7_POLICY_ACTION_REJECT})
            context.session.delete(pool_db)

    def get_pools(self, context, filters=None):
        pool_dbs = self._get_resources(context, models.PoolV2, filters=filters)
        return [data_models.Pool.from_sqlalchemy_model(pool_db)
                for pool_db in pool_dbs]

    def get_pools_as_api_dict(self, context, filters=None):
        options = (
            subqueryload(models.PoolV2.members),
            subqueryload(models.PoolV2.listeners),
            subqueryload(models.PoolV2.l7_policies),
            subqueryload(models.PoolV2.loadbalancer),
            subqueryload(models.PoolV2.session_persistence)
        )
        pool_dbs = self._get_resources(context, models.PoolV2,
            filters=filters, options=options)
        return [pool_db.to_api_dict
                for pool_db in pool_dbs]

    def get_pool(self, context, id):
        pool_db = self._get_resource(context, models.PoolV2, id)
        return data_models.Pool.from_sqlalchemy_model(pool_db)

    def create_pool_member(self, context, member, pool_id):
        try:
            with context.session.begin(subtransactions=True):
                self._load_id(context, member)
                member['pool_id'] = pool_id
                member['provisioning_status'] = n_const.PENDING_CREATE
                member['operating_status'] = lb_const.OFFLINE
                member_db = models.MemberV2(**member)
                context.session.add(member_db)
        except exception.DBDuplicateEntry:
            raise loadbalancerv2.MemberExists(address=member['address'],
                                              port=member['protocol_port'],
                                              pool=pool_id)
        context.session.refresh(member_db.pool)
        return data_models.Member.from_sqlalchemy_model(member_db)

    def update_pool_member(self, context, id, member):
        with context.session.begin(subtransactions=True):
            member_db = self._get_resource(context, models.MemberV2, id)
            member_db.update(member)
        context.session.refresh(member_db)
        return data_models.Member.from_sqlalchemy_model(member_db)

    def delete_pool_member(self, context, id):
        with context.session.begin(subtransactions=True):
            member_db = self._get_resource(context, models.MemberV2, id)
            context.session.delete(member_db)

    def get_pool_members(self, context, filters=None):
        filters = filters or {}
        member_dbs = self._get_resources(context, models.MemberV2,
                                         filters=filters)
        return [data_models.Member.from_sqlalchemy_model(member_db)
                for member_db in member_dbs]

    def get_pool_members_as_api_dict(self, context, filters=None):
        filters = filters or {}
        member_dbs = self._get_resources(context, models.MemberV2,
                                         filters=filters)
        return [member_db.to_api_dict
                for member_db in member_dbs]

    def get_pool_member(self, context, id):
        member_db = self._get_resource(context, models.MemberV2, id)
        return data_models.Member.from_sqlalchemy_model(member_db)

    def delete_member(self, context, id):
        with context.session.begin(subtransactions=True):
            member_db = self._get_resource(context, models.MemberV2, id)
            context.session.delete(member_db)

    def create_healthmonitor_on_pool(self, context, pool_id, healthmonitor):
        with context.session.begin(subtransactions=True):
            hm_db = self.create_healthmonitor(context, healthmonitor)
            pool = self.get_pool(context, pool_id)
            # do not want listener, members, l7_policies, healthmonitor
            # or loadbalancer in dict
            pool_dict = pool.to_dict(listeners=False, members=False,
                                     l7_policies=False, healthmonitor=False,
                                     loadbalancer=False, listener=False,
                                     loadbalancer_id=False)
            pool_dict['healthmonitor_id'] = hm_db.id
            self.update_pool(context, pool_id, pool_dict)
            hm_db = self._get_resource(context, models.HealthMonitorV2,
                                       hm_db.id)
        return data_models.HealthMonitor.from_sqlalchemy_model(hm_db)

    def create_healthmonitor(self, context, healthmonitor):
        with context.session.begin(subtransactions=True):
            self._load_id(context, healthmonitor)
            healthmonitor['provisioning_status'] = n_const.PENDING_CREATE
            hm_db_entry = models.HealthMonitorV2(**healthmonitor)
            context.session.add(hm_db_entry)
        return data_models.HealthMonitor.from_sqlalchemy_model(hm_db_entry)

    def update_healthmonitor(self, context, id, healthmonitor):
        with context.session.begin(subtransactions=True):
            hm_db = self._get_resource(context, models.HealthMonitorV2, id)
            hm_db.update(healthmonitor)
        context.session.refresh(hm_db)
        return data_models.HealthMonitor.from_sqlalchemy_model(hm_db)

    def delete_healthmonitor(self, context, id):
        with context.session.begin(subtransactions=True):
            hm_db_entry = self._get_resource(context,
                                             models.HealthMonitorV2, id)
            # TODO(sbalukoff): Clear out pool.healthmonitor_ids referencing
            # old healthmonitor ID.
            context.session.delete(hm_db_entry)

    def get_healthmonitor(self, context, id):
        hm_db = self._get_resource(context, models.HealthMonitorV2, id)
        return data_models.HealthMonitor.from_sqlalchemy_model(hm_db)

    def get_healthmonitors(self, context, filters=None):
        filters = filters or {}
        hm_dbs = self._get_resources(context, models.HealthMonitorV2,
                                     filters=filters)
        return [data_models.HealthMonitor.from_sqlalchemy_model(hm_db)
                for hm_db in hm_dbs]

    def get_healthmonitors_as_api_dict(self, context, filters=None):
        options = (
            subqueryload(models.HealthMonitorV2.pool)
        )
        filters = filters or {}
        hm_dbs = self._get_resources(context, models.HealthMonitorV2,
                                     filters=filters, options=options)
        return [hm_db.to_api_dict
                for hm_db in hm_dbs]

    def update_loadbalancer_stats(self, context, loadbalancer_id, stats_data):
        stats_data = stats_data or {}
        with context.session.begin(subtransactions=True):
            lb_db = self._get_resource(context, models.LoadBalancer,
                                       loadbalancer_id)
            lb_db.stats = self._create_loadbalancer_stats(context,
                                                          loadbalancer_id,
                                                          data=stats_data)

    def stats(self, context, loadbalancer_id):
        loadbalancer = self._get_resource(context, models.LoadBalancer,
                                          loadbalancer_id)
        return data_models.LoadBalancerStatistics.from_sqlalchemy_model(
            loadbalancer.stats)

    def create_l7policy(self, context, l7policy):
        if (l7policy.get('redirect_pool_id') and
                l7policy['redirect_pool_id'] == n_const.ATTR_NOT_SPECIFIED):
            l7policy['redirect_pool_id'] = None
        if not l7policy.get('position'):
            l7policy['position'] = 2147483647
        self._validate_l7policy_data(context, l7policy)

        with context.session.begin(subtransactions=True):
            listener_id = l7policy.get('listener_id')
            listener_db = self._get_resource(
                context, models.Listener, listener_id)

            self._load_id(context, l7policy)

            l7policy['provisioning_status'] = n_const.PENDING_CREATE

            l7policy_db = models.L7Policy(**l7policy)
            # MySQL int fields are by default 32-bit whereas handy system
            # constants like sys.maxsize are 64-bit on most platforms today.
            # Hence the reason this is 2147483647 (2^31 - 1) instead of an
            # elsewhere-defined constant.
            if l7policy['position'] == 2147483647:
                listener_db.l7_policies.append(l7policy_db)
            else:
                listener_db.l7_policies.insert(l7policy['position'] - 1,
                                               l7policy_db)

            listener_db.l7_policies.reorder()

        return data_models.L7Policy.from_sqlalchemy_model(l7policy_db)

    def update_l7policy(self, context, id, l7policy):
        with context.session.begin(subtransactions=True):

            l7policy_db = self._get_resource(context, models.L7Policy, id)

            if 'action' in l7policy:
                l7policy['listener_id'] = l7policy_db.listener_id
                self._validate_l7policy_data(context, l7policy)

            if ('position' not in l7policy or
                l7policy['position'] == 2147483647 or
                l7policy_db.position == l7policy['position']):
                l7policy_db.update(l7policy)
            else:
                listener_id = l7policy_db.listener_id
                listener_db = self._get_resource(
                    context, models.Listener, listener_id)
                l7policy_db = listener_db.l7_policies.pop(
                    l7policy_db.position - 1)
                # NOTE(ihrachys) create a new policy model because the one just
                # popped off the list became expired thanks to neutron db code
                # that automatically expires relationships when foreign keys
                # become obsolete
                l7policy_db = models.L7Policy(**dict(l7policy_db))
                l7policy_db.update(l7policy)
                listener_db.l7_policies.insert(l7policy['position'] - 1,
                                               l7policy_db)
                listener_db.l7_policies.reorder()

        context.session.refresh(l7policy_db)
        return data_models.L7Policy.from_sqlalchemy_model(l7policy_db)

    def delete_l7policy(self, context, id):
        with context.session.begin(subtransactions=True):
            l7policy_db = self._get_resource(context, models.L7Policy, id)
            listener_id = l7policy_db.listener_id
            listener_db = self._get_resource(
                context, models.Listener, listener_id)
            listener_db.l7_policies.remove(l7policy_db)

    def get_l7policy(self, context, id):
        l7policy_db = self._get_resource(context, models.L7Policy, id)
        return data_models.L7Policy.from_sqlalchemy_model(l7policy_db)

    def get_l7policies(self, context, filters=None):
        l7policy_dbs = self._get_resources(context, models.L7Policy,
                                           filters=filters)
        return [data_models.L7Policy.from_sqlalchemy_model(l7policy_db)
                for l7policy_db in l7policy_dbs]

    def get_l7policies_as_api_dict(self, context, filters=None):
        options = (
            subqueryload(models.L7Policy.rules)
        )
        l7policy_dbs = self._get_resources(context, models.L7Policy,
                                           filters=filters, options=options)
        return [l7policy_db.to_api_dict
                for l7policy_db in l7policy_dbs]

    def create_l7policy_rule(self, context, rule, l7policy_id):
        with context.session.begin(subtransactions=True):
            if not self._resource_exists(context, models.L7Policy,
                                         l7policy_id):
                raise loadbalancerv2.EntityNotFound(
                    name=models.L7Policy.NAME, id=l7policy_id)
            self._validate_l7rule_data(context, rule)
            self._load_id(context, rule)
            rule['l7policy_id'] = l7policy_id
            rule['provisioning_status'] = n_const.PENDING_CREATE
            rule_db = models.L7Rule(**rule)
            context.session.add(rule_db)
        return data_models.L7Rule.from_sqlalchemy_model(rule_db)

    def update_l7policy_rule(self, context, id, rule, l7policy_id):
        with context.session.begin(subtransactions=True):
            if not self._resource_exists(context, models.L7Policy,
                                         l7policy_id):
                raise l7.RuleNotFoundForL7Policy(
                    l7policy_id=l7policy_id, rule_id=id)

            rule_db = self._get_resource(context, models.L7Rule, id)
            # If user did not intend to change all parameters,
            # already stored parameters will be used for validations
            if not rule.get('type'):
                rule['type'] = rule_db.type
            if not rule.get('value'):
                rule['value'] = rule_db.value
            if not rule.get('compare_type'):
                rule['compare_type'] = rule_db.compare_type

            self._validate_l7rule_data(context, rule)
            rule_db = self._get_resource(context, models.L7Rule, id)
            rule_db.update(rule)
        context.session.refresh(rule_db)
        return data_models.L7Rule.from_sqlalchemy_model(rule_db)

    def delete_l7policy_rule(self, context, id):
        with context.session.begin(subtransactions=True):
            rule_db_entry = self._get_resource(context, models.L7Rule, id)
            context.session.delete(rule_db_entry)

    def get_l7policy_rule(self, context, id, l7policy_id):
        rule_db = self._get_resource(context, models.L7Rule, id)
        if rule_db.l7policy_id != l7policy_id:
            raise l7.RuleNotFoundForL7Policy(
                l7policy_id=l7policy_id, rule_id=id)
        return data_models.L7Rule.from_sqlalchemy_model(rule_db)

    def get_l7policy_rules(self, context, l7policy_id, filters=None):
        if filters:
            filters.update(filters)
        else:
            filters = {'l7policy_id': [l7policy_id]}
        rule_dbs = self._get_resources(context, models.L7Rule,
                                       filters=filters)
        return [data_models.L7Rule.from_sqlalchemy_model(rule_db)
                for rule_db in rule_dbs]

    def get_l7policy_rules_as_api_dict(
            self, context, l7policy_id, filters=None):
        options = (
            subqueryload(models.L7Rule.policy)
        )
        if filters:
            filters.update(filters)
        else:
            filters = {'l7policy_id': [l7policy_id]}
        rule_dbs = self._get_resources(context, models.L7Rule,
                                       filters=filters, options=options)
        return [rule_db.to_api_dict
                for rule_db in rule_dbs]


def _prevent_lbaasv2_port_delete_callback(resource, event, trigger, **kwargs):
    context = kwargs['context']
    port_id = kwargs['port_id']
    port_check = kwargs['port_check']
    lbaasv2plugin = directory.get_plugin(pg_const.LOADBALANCERV2)
    if lbaasv2plugin and port_check:
        lbaasv2plugin.db.prevent_lbaasv2_port_deletion(context, port_id)



# Copyright 2014-2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add provisioning and operating statuses

Revision ID: 4deef6d81931
Revises: lbaasv2
Create Date: 2015-01-27 20:38:20.796401

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4deef6d81931'
down_revision = 'lbaasv2'

PROVISIONING_STATUS = u'provisioning_status'
OPERATING_STATUS = u'operating_status'
STATUS = u'status'


def upgrade():
    op.drop_column(u'lbaas_loadbalancers', STATUS)
    op.add_column(
        u'lbaas_loadbalancers',
        sa.Column(PROVISIONING_STATUS, sa.String(16), nullable=False)
    )
    op.add_column(
        u'lbaas_loadbalancers',
        sa.Column(OPERATING_STATUS, sa.String(16), nullable=False)
    )
    op.drop_column(u'lbaas_listeners', STATUS)
    op.add_column(
        u'lbaas_listeners',
        sa.Column(PROVISIONING_STATUS, sa.String(16), nullable=False)
    )
    op.add_column(
        u'lbaas_listeners',
        sa.Column(OPERATING_STATUS, sa.String(16), nullable=False)
    )
    op.drop_column(u'lbaas_pools', STATUS)
    op.add_column(
        u'lbaas_pools',
        sa.Column(PROVISIONING_STATUS, sa.String(16), nullable=False)
    )
    op.add_column(
        u'lbaas_pools',
        sa.Column(OPERATING_STATUS, sa.String(16), nullable=False)
    )
    op.drop_column(u'lbaas_members', STATUS)
    op.add_column(
        u'lbaas_members',
        sa.Column(PROVISIONING_STATUS, sa.String(16), nullable=False)
    )
    op.add_column(
        u'lbaas_members',
        sa.Column(OPERATING_STATUS, sa.String(16), nullable=False)
    )
    op.drop_column(u'lbaas_healthmonitors', STATUS)
    op.add_column(
        u'lbaas_healthmonitors',
        sa.Column(PROVISIONING_STATUS, sa.String(16), nullable=False)
    )



# Copyright 2017, Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from neutron.db import servicetype_db as st_db
from neutron.services import provider_configuration as pconf
from neutron_lib import exceptions as lib_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import requests

from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.services.loadbalancer import constants

LOG = logging.getLogger(__name__)
VERSION = 'v2.0'
OCTAVIA_PROXY_CLIENT = (
    "LBaaS V2 Octavia Proxy/{version} "
    "(https://wiki.openstack.org/wiki/Octavia)").format(version=VERSION)
FILTER = ['vip_address', 'vip_network_id', 'flavor_id',
          'provider', 'redirect_pool_id']

LOADBALANCER = 'loadbalancer'
LISTENER = 'listener'
POOL = 'pool'
L7POLICY = 'l7policy'
L7POLICY_RULE = 'rule'
MEMBER = 'member'
HEALTH_MONITOR = 'healthmonitor'
STATUS = 'status'
GRAPH = 'graph'
STATS = 'stats'

OPTS = [
    cfg.StrOpt(
        'base_url',
        default='http://127.0.0.1:9876',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('URL of Octavia controller root'),
    ),
]

cfg.CONF.register_opts(OPTS, 'octavia')


def add_provider_configuration(type_manager, service_type):
    type_manager.add_provider_configuration(
        service_type,
        pconf.ProviderConfiguration('neutron_lbaas_proxy'))


class LoadBalancerProxyPluginv2(loadbalancerv2.LoadBalancerPluginBaseV2):
    """Implementation of the Neutron Loadbalancer Proxy Plugin.

    This class proxies all requests/reponses to Octavia
    """

    supported_extension_aliases = ["lbaasv2",
                                   "shared_pools",
                                   "l7",
                                   "lbaas_agent_schedulerv2",
                                   "service-type",
                                   "lb-graph",
                                   "lb_network_vip",
                                   "hm_max_retries_down"]
    path_prefix = loadbalancerv2.LOADBALANCERV2_PREFIX

    def __init__(self):
        LOG.warning('neutron-lbaas is now deprecated. See: '
                    'https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                    'Deprecation')
        self.service_type_manager = st_db.ServiceTypeManager.get_instance()
        add_provider_configuration(
            self.service_type_manager, constants.LOADBALANCERV2)
        self.get = functools.partial(self.request, 'GET')
        self.post = functools.partial(self.request, 'POST')
        self.put = functools.partial(self.request, 'PUT')
        self.delete = functools.partial(self.request, 'DELETE')
        self.base_url = '{}/{}/lbaas'.format(cfg.CONF.octavia.base_url,
                                             VERSION)

    def get_plugin_type(self):
        return constants.LOADBALANCERV2

    def get_plugin_description(self):
        return "Neutron LoadBalancer Proxy Plugin"

    def request(self, method, url, token=None, args=None, headers=None,
                accepted_codes=[200, 201, 202, 204]):
        params = {}
        if args:
            # extract filter and fields
            if 'filters' in args:
                params = args.pop('filters')
            if 'fields' in args:
                params['fields'] = args.pop('fields')
            args = jsonutils.dumps(args)
        if not headers:
            headers = {
                'Content-type': 'application/json',
                'X-Auth-Token': token
            }
        headers['User-Agent'] = OCTAVIA_PROXY_CLIENT

        url = '{}/{}'.format(self.base_url, str(url))
        LOG.debug("url = %s", url)
        LOG.debug("args = %s", args)
        LOG.debug("params = %s", str(params))
        r = requests.request(method, url, data=args, params=params,
                             headers=headers)
        LOG.debug("Octavia Response Code: {0}".format(r.status_code))
        LOG.debug("Octavia Response Body: {0}".format(r.content))
        LOG.debug("Octavia Response Headers: {0}".format(r.headers))

        if r.status_code in accepted_codes:
            if method != 'DELETE':
                return r.json()
        elif r.status_code == 413:
            e = lib_exc.OverQuota()
            e.msg = str(r.content)
            raise e
        elif r.status_code == 409:
            e = lib_exc.Conflict()
            e.msg = str(r.content)
            raise e
        elif r.status_code == 401:
            e = lib_exc.NotAuthorized()
            e.msg = str(r.content)
            raise e
        elif r.status_code == 404:
            e = lib_exc.NotFound()
            e.msg = str(r.content)
            raise e
        elif r.status_code == 400:
            e = lib_exc.BadRequest(resource="", msg="")
            e.msg = str(r.content)
            raise e
        else:
            raise loadbalancerv2.DriverError(msg=str(r.content))

    def _filter(self, keys, map):
        """Filter the args map

        keys: The keys to filter out
        map: the args in a map

        NOTE: This returns a deep copy - leaving the original alone
        """

        res = {}
        for k in map:
            if k not in keys:
                if map[k]:
                    res[k] = map[k]
        if 'tenant_id' in res:
            res['project_id'] = res.pop('tenant_id')
        return res

    def pluralize(self, name):
        if name.endswith('y'):
            return name[:-1] + "ies"
        elif not name.endswith('s'):
            return "{}s".format(name)
        return name

    def _path(self, resource, sub_resource, resource_id):
        url = resource
        if sub_resource:
            url = "{}/{}/{}".format(self.pluralize(resource),
                                    resource_id, sub_resource)
        return self.pluralize(url)

    def _create_resource(self, resource, context, res, sub_resource=None,
                         resource_id=None):
        # clean up the map
        resource_ = resource if not sub_resource else sub_resource
        r = self._filter(FILTER, res[resource_])
        r = self.post(self._path(resource, sub_resource, resource_id),
                      context.auth_token, {resource_: r})
        return r[resource_]

    def _get_resources(self, resource, context, filters=None, fields=None,
                       sub_resource=None, resource_id=None,
                       pass_through=False):
        # not sure how to test that or if we even support sorting/filtering?
        resource_ = resource if not sub_resource else sub_resource
        args = {}
        if filters:
            if 'tenant_id' in filters:
                filters['project_id'] = filters.pop('tenant_id')
            args['filters'] = filters
        if fields:
            args['fields'] = fields
        res = self.get(self._path(resource, sub_resource, resource_id),
                       context.auth_token, args)
        return res[self.pluralize(resource_)] if not pass_through else res

    def _get_resource(self, resource, context, id, fields=None,
                      sub_resource=None, resource_id=None):
        # not sure how to test that or if we even support sorting/filtering?
        args = {}
        if fields:
            args['fields'] = fields
        resource_ = resource if not sub_resource else sub_resource
        res = self.get('{}/{}'.format(
            self._path(resource, sub_resource, resource_id), id),
                       context.auth_token, args)
        return res[resource_]

    def _update_resource(self, resource, context, id, res,
                         sub_resource=None, resource_id=None):
        # clean up the map
        resource_ = resource if not sub_resource else sub_resource
        r = self._filter(FILTER, res[resource_])
        res = self.put('{}/{}'.format(self._path(
            resource, sub_resource, resource_id), id),
                       context.auth_token,
                       {resource_: r})
        return res[resource_]

    def _delete_resource(self, resource, context, id,
                         sub_resource=None, resource_id=None):
        self.delete('{}/{}'.format(self._path(
            resource, sub_resource, resource_id), id),
            context.auth_token)

    def create_loadbalancer(self, context, loadbalancer):
        return self._create_resource(LOADBALANCER, context, loadbalancer)

    def get_loadbalancers(self, context, filters=None, fields=None):
        return self._get_resources(LOADBALANCER, context, filters, fields)

    def get_loadbalancer(self, context, id, fields=None):
        return self._get_resource(LOADBALANCER, context, id, fields)

    def update_loadbalancer(self, context, id, loadbalancer):
        return self._update_resource(LOADBALANCER, context, id, loadbalancer)

    def delete_loadbalancer(self, context, id):
        self._delete_resource(LOADBALANCER, context, id)

    def create_listener(self, context, listener):
        return self._create_resource(LISTENER, context, listener)

    def get_listener(self, context, id, fields=None):
        return self._get_resource(LISTENER, context, id, fields)

    def get_listeners(self, context, filters=None, fields=None):
        return self._get_resources(LISTENER, context, filters, fields)

    def update_listener(self, context, id, listener):
        return self._update_resource(LISTENER, context, id, listener)

    def delete_listener(self, context, id):
        return self._delete_resource(LISTENER, context, id)

    def get_pools(self, context, filters=None, fields=None):
        return self._get_resources(POOL, context, filters, fields)

    def get_pool(self, context, id, fields=None):
        return self._get_resource(POOL, context, id, fields)

    def create_pool(self, context, pool):
        return self._create_resource(POOL, context, pool)

    def update_pool(self, context, id, pool):
        return self._update_resource(POOL, context, id, pool)

    def delete_pool(self, context, id):
        return self._delete_resource(POOL, context, id)

    def get_pool_members(self, context, pool_id,
                         filters=None,
                         fields=None):
        return self._get_resources(POOL, context, filters, fields,
                                   MEMBER, pool_id)

    def get_pool_member(self, context, id, pool_id,
                        fields=None):
        return self._get_resource(POOL, context, id, fields,
                                  MEMBER, pool_id)

    def create_pool_member(self, context, pool_id, member):
        return self._create_resource(POOL, context, member, MEMBER, pool_id)

    def update_pool_member(self, context, id, pool_id, member):
        return self._update_resource(POOL, context, id, member,
                                     MEMBER, pool_id)

    def delete_pool_member(self, context, id, pool_id):
        return self._delete_resource(POOL, context, id, MEMBER, pool_id)

    def get_healthmonitors(self, context, filters=None, fields=None):
        return self._get_resources(HEALTH_MONITOR, context, filters, fields)

    def get_healthmonitor(self, context, id, fields=None):
        return self._get_resource(HEALTH_MONITOR, context, id, fields)

    def create_healthmonitor(self, context, healthmonitor):
        return self._create_resource(HEALTH_MONITOR, context, healthmonitor)

    def update_healthmonitor(self, context, id, healthmonitor):
        return self._update_resource(HEALTH_MONITOR, context,
                                     id, healthmonitor)

    def delete_healthmonitor(self, context, id):
        return self._delete_resource(HEALTH_MONITOR, context, id)

    def get_members(self, context, filters=None, fields=None):
        pass

    def get_member(self, context, id, fields=None):
        pass

    def statuses(self, context, loadbalancer_id):
        return self._get_resources(LOADBALANCER, context, sub_resource=STATUS,
                                   resource_id=loadbalancer_id,
                                   pass_through=True)

    def get_l7policies(self, context, filters=None, fields=None):
        return self._get_resources(L7POLICY, context, filters, fields)

    def get_l7policy(self, context, id, fields=None):
        return self._get_resource(L7POLICY, context, id, fields)

    def create_l7policy(self, context, l7policy):
        return self._create_resource(L7POLICY, context, l7policy)

    def update_l7policy(self, context, id, l7policy):
        return self._update_resource(L7POLICY, context, id, l7policy)

    def delete_l7policy(self, context, id):
        return self._delete_resource(L7POLICY, context, id)

    def get_l7policy_rules(self, context, l7policy_id,
                           filters=None, fields=None):
        return self._get_resources(L7POLICY, context, filters, fields,
                                   L7POLICY_RULE, l7policy_id)

    def get_l7policy_rule(self, context, id, l7policy_id, fields=None):
        return self._get_resource(L7POLICY, context, id, fields,
                                  L7POLICY_RULE, l7policy_id)

    def create_l7policy_rule(self, context, rule, l7policy_id):
        return self._create_resource(L7POLICY, context, rule, L7POLICY_RULE,
                                     l7policy_id)

    def update_l7policy_rule(self, context, id, rule, l7policy_id):
        return self._update_resource(L7POLICY, context, id, rule,
                                     L7POLICY_RULE, l7policy_id)

    def delete_l7policy_rule(self, context, id, l7policy_id):
        return self._delete_resource(L7POLICY, context, id, L7POLICY_RULE,
                                     l7policy_id)

    def create_graph(self, context, graph):
        return self._create_resource(GRAPH, context, graph)

    def stats(self, context, loadbalancer_id):
        return self._get_resources(LOADBALANCER, context, sub_resource=STATS,
                                   resource_id=loadbalancer_id,
                                   pass_through=True)



# Copyright (c) 2013 OpenStack Foundation.
# Copyright 2015 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from datetime import datetime

import mock
from neutron_lib import context
from neutron_lib.plugins import directory

from neutron.api import extensions
from neutron.db import agents_db
from neutron.tests.common import helpers
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.db import test_agentschedulers_db
import neutron.tests.unit.extensions
from neutron.tests.unit.extensions import test_agent
from neutron_lib import constants as n_constants
from neutron_lib.plugins import constants as plugin_const
from webob import exc

from neutron_lbaas.drivers.haproxy import plugin_driver
from neutron_lbaas.extensions import lbaas_agentschedulerv2
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.tests import base
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2
from neutron_lbaas.tests.unit.db.loadbalancer import util

LBAAS_HOSTA = 'hosta'
extensions_path = ':'.join(neutron.tests.unit.extensions.__path__)


class AgentSchedulerTestMixIn(test_agentschedulers_db.AgentSchedulerTestMixIn):
    def _list_loadbalancers_hosted_by_agent(
            self, agent_id, expected_code=exc.HTTPOk.code, admin_context=True):
        path = "/agents/%s/%s.%s" % (agent_id,
                                     lbaas_agentschedulerv2.LOADBALANCERS,
                                     self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _get_lbaas_agent_hosting_loadbalancer(self, loadbalancer_id,
                                              expected_code=exc.HTTPOk.code,
                                              admin_context=True):
        path = "/lbaas/loadbalancers/%s/%s.%s" % (loadbalancer_id,
                                                  lbaas_agentschedulerv2
                                                  .LOADBALANCER_AGENT,
                                                  self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)


class LBaaSAgentSchedulerTestCase(test_agent.AgentDBTestMixIn,
                                  AgentSchedulerTestMixIn,
                                  util.LbaasTestMixin,
                                  base.NeutronDbPluginV2TestCase):
    fmt = 'json'
    plugin_str = 'neutron.plugins.ml2.plugin.Ml2Plugin'

    def _register_agent_states(self, lbaas_agents=False):
        res = super(LBaaSAgentSchedulerTestCase, self)._register_agent_states(
            lbaas_agents=lbaas_agents)
        if lbaas_agents:
            lbaas_hosta = {
                'binary': 'neutron-loadbalancer-agent',
                'host': test_agent.LBAAS_HOSTA,
                'topic': 'LOADBALANCER_AGENT',
                'configurations': {'device_drivers': [
                    plugin_driver.HaproxyOnHostPluginDriver.device_driver]},
                'agent_type': lb_const.AGENT_TYPE_LOADBALANCERV2}
            lbaas_hostb = copy.deepcopy(lbaas_hosta)
            lbaas_hostb['host'] = test_agent.LBAAS_HOSTB
            callback = agents_db.AgentExtRpcCallback()
            callback.report_state(self.adminContext,
                                  agent_state={'agent_state': lbaas_hosta},
                                  time=datetime.utcnow().isoformat())
            callback.report_state(self.adminContext,
                                  agent_state={'agent_state': lbaas_hostb},
                                  time=datetime.utcnow().isoformat())
            res += [lbaas_hosta, lbaas_hostb]
        return res

    def setUp(self):
        service_plugins = {
            'lb_plugin_name': test_db_loadbalancerv2.DB_LB_PLUGIN_CLASS}

        # default provider should support agent scheduling
        self.set_override(
            [('LOADBALANCERV2:lbaas:neutron_lbaas.drivers.haproxy.'
              'plugin_driver.HaproxyOnHostPluginDriver:default')])

        super(LBaaSAgentSchedulerTestCase, self).setUp(
            self.plugin_str, service_plugins=service_plugins)
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.adminContext = context.get_admin_context()
        self.lbaas_plugin = directory.get_plugin(plugin_const.LOADBALANCERV2)
        self.core_plugin = directory.get_plugin()

    def test_report_states(self):
        self._register_agent_states(lbaas_agents=True)
        agents = self._list_agents()
        self.assertEqual(8, len(agents['agents']))

    def test_loadbalancer_scheduling_on_loadbalancer_creation(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            lbaas_agent = self._get_lbaas_agent_hosting_loadbalancer(
                loadbalancer['loadbalancer']['id'])
            self.assertIsNotNone(lbaas_agent)
            self.assertEqual(lb_const.AGENT_TYPE_LOADBALANCERV2,
                             lbaas_agent['agent']['agent_type'])
            loadbalancers = self._list_loadbalancers_hosted_by_agent(
                lbaas_agent['agent']['id'])
            self.assertEqual(1, len(loadbalancers['loadbalancers']))
            self.assertEqual(loadbalancer['loadbalancer'],
                             loadbalancers['loadbalancers'][0])
            self.lbaas_plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer['loadbalancer']['id']
            )

    def test_schedule_loadbalancer_with_disabled_agent(self):
        lbaas_hosta = {
            'binary': 'neutron-loadbalancer-agent',
            'host': LBAAS_HOSTA,
            'topic': 'LOADBALANCER_AGENT',
            'configurations': {'device_drivers': [
                plugin_driver.HaproxyOnHostPluginDriver.device_driver
            ]},
            'agent_type': lb_const.AGENT_TYPE_LOADBALANCERV2}
        helpers._register_agent(lbaas_hosta)
        with self.loadbalancer() as loadbalancer:
            lbaas_agent = self._get_lbaas_agent_hosting_loadbalancer(
                loadbalancer['loadbalancer']['id'])
            self.assertIsNotNone(lbaas_agent)
            self.lbaas_plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer['loadbalancer']['id']
            )
        agents = self._list_agents()
        self._disable_agent(agents['agents'][0]['id'])
        subnet = self.core_plugin.get_subnets(self.adminContext)[0]
        lb = {
            'loadbalancer': {
                'vip_subnet_id': subnet['id'],
                'provider': 'lbaas',
                'flavor_id': n_constants.ATTR_NOT_SPECIFIED,
                'vip_address': n_constants.ATTR_NOT_SPECIFIED,
                'admin_state_up': True,
                'tenant_id': self._tenant_id,
                'listeners': []}}
        self.assertRaises(lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                          self.lbaas_plugin.create_loadbalancer,
                          self.adminContext, lb)

    def test_schedule_loadbalancer_with_down_agent(self):
        lbaas_hosta = {
            'binary': 'neutron-loadbalancer-agent',
            'host': LBAAS_HOSTA,
            'topic': 'LOADBALANCER_AGENT',
            'configurations': {'device_drivers': [
                plugin_driver.HaproxyOnHostPluginDriver.device_driver
            ]},
            'agent_type': lb_const.AGENT_TYPE_LOADBALANCERV2}
        helpers._register_agent(lbaas_hosta)
        is_agent_down_str = 'neutron.agent.common.utils.is_agent_down'
        with mock.patch(is_agent_down_str) as mock_is_agent_down:
            mock_is_agent_down.return_value = False
            with self.loadbalancer() as loadbalancer:
                lbaas_agent = self._get_lbaas_agent_hosting_loadbalancer(
                    loadbalancer['loadbalancer']['id'])
                self.lbaas_plugin.db.update_loadbalancer_provisioning_status(
                    self.adminContext, loadbalancer['loadbalancer']['id']
                )
            self.assertIsNotNone(lbaas_agent)
        with mock.patch(is_agent_down_str) as mock_is_agent_down:
            mock_is_agent_down.return_value = True
            subnet = self.core_plugin.get_subnets(self.adminContext)[0]
            lb = {
                'loadbalancer': {
                    'vip_subnet_id': subnet['id'],
                    'provider': 'lbaas',
                    'flavor_id': n_constants.ATTR_NOT_SPECIFIED,
                    'vip_address': n_constants.ATTR_NOT_SPECIFIED,
                    'admin_state_up': True,
                    'tenant_id': self._tenant_id,
                    'listeners': []}}
            self.assertRaises(lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                              self.lbaas_plugin.create_loadbalancer,
                              self.adminContext, lb)

    def test_loadbalancer_unscheduling_on_loadbalancer_deletion(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            lbaas_agent = self._get_lbaas_agent_hosting_loadbalancer(lb_id)
            self.assertIsNotNone(lbaas_agent)
            self.assertEqual(lb_const.AGENT_TYPE_LOADBALANCERV2,
                             lbaas_agent['agent']['agent_type'])
            loadbalancers = self._list_loadbalancers_hosted_by_agent(
                lbaas_agent['agent']['id'])
            self.assertEqual(1, len(loadbalancers['loadbalancers']))
            self.assertEqual(loadbalancer['loadbalancer'],
                             loadbalancers['loadbalancers'][0])

            self.lbaas_plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, lb_id
            )

            req = self.new_delete_request('loadbalancers', lb_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(exc.HTTPNoContent.code, res.status_int)
            loadbalancers = self._list_loadbalancers_hosted_by_agent(
                lbaas_agent['agent']['id'])
            self.assertEqual(0, len(loadbalancers['loadbalancers']))

    def test_loadbalancer_scheduling_non_admin_access(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            self._get_lbaas_agent_hosting_loadbalancer(
                loadbalancer['loadbalancer']['id'],
                expected_code=exc.HTTPForbidden.code,
                admin_context=False)
            self._list_loadbalancers_hosted_by_agent(
                'fake_id',
                expected_code=exc.HTTPForbidden.code,
                admin_context=False)
            self.lbaas_plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer['loadbalancer']['id']
            )



# Copyright 2015 Rackspace.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
import oslo_messaging as messaging

from neutron_lbaas._i18n import _
from neutron_lbaas.db.loadbalancer import loadbalancer_dbv2
from neutron_lbaas.db.loadbalancer import models as db_models
from neutron_lbaas.services.loadbalancer import data_models

LOG = logging.getLogger(__name__)


class LoadBalancerCallbacks(object):

    # history
    #   1.0 Initial version
    target = messaging.Target(version='1.0')

    def __init__(self, plugin):
        super(LoadBalancerCallbacks, self).__init__()
        self.plugin = plugin

    def get_ready_devices(self, context, host=None):
        with context.session.begin(subtransactions=True):
            agents = self.plugin.db.get_lbaas_agents(
                context, filters={'host': [host]})
            if not agents:
                return []
            elif len(agents) > 1:
                LOG.warning('Multiple lbaas agents found on host %s', host)
            loadbalancers = self.plugin.db.list_loadbalancers_on_lbaas_agent(
                context, agents[0].id)
            loadbalancer_ids = [
                l.id for l in loadbalancers]

            qry = context.session.query(
                loadbalancer_dbv2.models.LoadBalancer.id)
            qry = qry.filter(
                loadbalancer_dbv2.models.LoadBalancer.id.in_(
                    loadbalancer_ids))
            qry = qry.filter(
                loadbalancer_dbv2.models.LoadBalancer.provisioning_status.in_(
                    constants.ACTIVE_PENDING_STATUSES))
            up = True  # makes pep8 and sqlalchemy happy
            qry = qry.filter(
                loadbalancer_dbv2.models.LoadBalancer.admin_state_up == up)
            return [id for id, in qry]

    def get_loadbalancer(self, context, loadbalancer_id=None):
        lb_model = self.plugin.db.get_loadbalancer(context, loadbalancer_id)
        if lb_model.vip_port and lb_model.vip_port.fixed_ips:
            for fixed_ip in lb_model.vip_port.fixed_ips:
                subnet_dict = self.plugin.db._core_plugin.get_subnet(
                    context, fixed_ip.subnet_id
                )
                setattr(fixed_ip, 'subnet', data_models.Subnet.from_dict(
                    subnet_dict))
        if lb_model.provider:
            device_driver = self.plugin.drivers[
                lb_model.provider.provider_name].device_driver
            setattr(lb_model.provider, 'device_driver', device_driver)
        if lb_model.vip_port:
            network_dict = self.plugin.db._core_plugin.get_network(
                context, lb_model.vip_port.network_id)
            setattr(lb_model.vip_port, 'network',
                    data_models.Network.from_dict(network_dict))
        lb_dict = lb_model.to_dict(stats=False)

        return lb_dict

    def loadbalancer_deployed(self, context, loadbalancer_id):
        with context.session.begin(subtransactions=True):
            qry = context.session.query(db_models.LoadBalancer)
            qry = qry.filter_by(id=loadbalancer_id)
            loadbalancer = qry.one()

            # set all resources to active
            if (loadbalancer.provisioning_status in
                    constants.ACTIVE_PENDING_STATUSES):
                loadbalancer.provisioning_status = constants.ACTIVE

            if loadbalancer.listeners:
                for l in loadbalancer.listeners:
                    if (l.provisioning_status in
                            constants.ACTIVE_PENDING_STATUSES):
                        l.provisioning_status = constants.ACTIVE
                    if (l.default_pool and
                        l.default_pool.provisioning_status in
                            constants.ACTIVE_PENDING_STATUSES):
                        l.default_pool.provisioning_status = constants.ACTIVE
                        if l.default_pool.members:
                            for m in l.default_pool.members:
                                if (m.provisioning_status in
                                        constants.ACTIVE_PENDING_STATUSES):
                                    m.provisioning_status = constants.ACTIVE
                        if l.default_pool.healthmonitor:
                            hm = l.default_pool.healthmonitor
                            ps = hm.provisioning_status
                            if ps in constants.ACTIVE_PENDING_STATUSES:
                                (l.default_pool.healthmonitor
                                 .provisioning_status) = constants.ACTIVE

    def update_status(self, context, obj_type, obj_id,
                      provisioning_status=None, operating_status=None):
        if not provisioning_status and not operating_status:
            LOG.warning('update_status for %(obj_type)s %(obj_id)s called '
                        'without specifying provisioning_status or '
                        'operating_status' % {'obj_type': obj_type,
                                              'obj_id': obj_id})
            return
        model_mapping = {
            'loadbalancer': db_models.LoadBalancer,
            'pool': db_models.PoolV2,
            'listener': db_models.Listener,
            'member': db_models.MemberV2,
            'healthmonitor': db_models.HealthMonitorV2
        }
        if obj_type not in model_mapping:
            raise n_exc.Invalid(_('Unknown object type: %s') % obj_type)
        try:
            self.plugin.db.update_status(
                context, model_mapping[obj_type], obj_id,
                provisioning_status=provisioning_status,
                operating_status=operating_status)
        except n_exc.NotFound:
            # update_status may come from agent on an object which was
            # already deleted from db with other request
            LOG.warning('Cannot update status: %(obj_type)s %(obj_id)s '
                        'not found in the DB, it was probably deleted '
                        'concurrently',
                        {'obj_type': obj_type, 'obj_id': obj_id})

    def loadbalancer_destroyed(self, context, loadbalancer_id=None):
        """Agent confirmation hook that a load balancer has been destroyed.

        This method exists for subclasses to change the deletion
        behavior.
        """
        pass

    def plug_vip_port(self, context, port_id=None, host=None):
        if not port_id:
            return

        try:
            port = self.plugin.db._core_plugin.get_port(
                context,
                port_id
            )
        except n_exc.PortNotFound:
            LOG.debug('Unable to find port %s to plug.', port_id)
            return

        port['admin_state_up'] = True
        port[portbindings.HOST_ID] = host
        port['device_owner'] = constants.DEVICE_OWNER_LOADBALANCERV2
        self.plugin.db._core_plugin.update_port(
            context,
            port_id,
            {'port': port}
        )

    def unplug_vip_port(self, context, port_id=None, host=None):
        if not port_id:
            return

        try:
            port = self.plugin.db._core_plugin.get_port(
                context,
                port_id
            )
        except n_exc.PortNotFound:
            LOG.debug('Unable to find port %s to unplug. This can occur when '
                      'the Vip has been deleted first.',
                      port_id)
            return

        port['admin_state_up'] = False
        port['device_owner'] = ''
        port['device_id'] = ''

        try:
            self.plugin.db._core_plugin.update_port(
                context,
                port_id,
                {'port': port}
            )

        except n_exc.PortNotFound:
            LOG.debug('Unable to find port %s to unplug.  This can occur when '
                      'the Vip has been deleted first.',
                      port_id)

    def update_loadbalancer_stats(self, context,
                                  loadbalancer_id=None,
                                  stats=None):
        self.plugin.db.update_loadbalancer_stats(context, loadbalancer_id,
                                                 stats)



# Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg
from oslo_messaging.rpc import dispatcher

from neutron_lbaas.common import exceptions
from neutron_lbaas.db.loadbalancer import models
import neutron_lbaas.drivers.octavia.driver as odriver
from neutron_lbaas.drivers.octavia.driver import octavia_messaging_consumer
from neutron_lbaas.services.loadbalancer import constants
from neutron_lbaas.tests.unit.drivers.octavia import test_octavia_driver

InfoContainer = octavia_messaging_consumer.InfoContainer


class TestOctaviaMessagingConsumer(test_octavia_driver.BaseOctaviaDriverTest):
    def setUp(self):
        super(test_octavia_driver.BaseOctaviaDriverTest, self).setUp()
        self.plugin = mock.Mock()
        self.driver = odriver.OctaviaDriver(self.plugin)

    def assert_handle_streamed_event_called(self, model_class, id_param,
                                            payload):
        call_args_list = self.driver.plugin.db.update_status.call_args_list[0]
        self.assertEqual(len(call_args_list), 2)
        self.assertEqual(len(call_args_list[0]), 3)
        self.assertEqual(model_class, call_args_list[0][1])
        self.assertEqual(call_args_list[0][2], id_param)
        self.assertEqual(call_args_list[1], payload)

    def test_info_container_constructor(self):
        ID = 'test_id'
        PAYLOAD = 'test_payload'
        TYPE = 'test_type'
        cnt = InfoContainer(TYPE, ID, PAYLOAD)
        self.assertEqual(cnt.info_type, TYPE)
        self.assertEqual(cnt.info_id, ID)
        self.assertEqual(cnt.info_payload, PAYLOAD)
        self.assertEqual(cnt.to_dict(), {'info_type': TYPE, 'info_id': ID,
        'info_payload': PAYLOAD})

    def test_info_container_from_dict(self):
        ID = 'test_id'
        PAYLOAD = 'test_payload'
        TYPE = 'test_type'
        cnt = InfoContainer.from_dict({'info_type': TYPE, 'info_id': ID,
        'info_payload': PAYLOAD})
        self.assertEqual(cnt.info_type, TYPE)
        self.assertEqual(cnt.info_id, ID)
        self.assertEqual(cnt.info_payload, PAYLOAD)

    def test_set_consumer_topic(self):
        TOPIC = 'neutron_lbaas_event'
        self.addCleanup(cfg.CONF.clear_override, 'event_stream_topic',
                        group='oslo_messaging')
        cfg.CONF.set_override('event_stream_topic', TOPIC,
                              group='oslo_messaging')
        consumer = octavia_messaging_consumer.OctaviaConsumer(self.driver)
        self.assertIsNotNone(consumer.transport)
        self.assertEqual(TOPIC, consumer.target.topic)
        self.assertEqual(cfg.CONF.host, consumer.target.server)

    @mock.patch.object(octavia_messaging_consumer.messaging, 'get_rpc_server')
    def test_consumer_start(self, mock_get_rpc_server):
        mock_server = mock.Mock()
        mock_get_rpc_server.return_value = mock_server
        TOPIC = 'neutron_lbaas_event'
        self.addCleanup(cfg.CONF.clear_override, 'event_stream_topic',
                        group='oslo_messaging')
        cfg.CONF.set_override('event_stream_topic', TOPIC,
                              group='oslo_messaging')
        consumer = octavia_messaging_consumer.OctaviaConsumer(self.driver)
        consumer.start()
        access_policy = dispatcher.DefaultRPCAccessPolicy
        mock_get_rpc_server.assert_called_once_with(
            consumer.transport, consumer.target, consumer.endpoints,
            executor='eventlet', access_policy=access_policy
        )
        mock_server.start.assert_called_once_with()

    @mock.patch.object(octavia_messaging_consumer.messaging, 'get_rpc_server')
    def test_consumer_stop(self, mock_get_rpc_server):
        mock_server = mock.Mock()
        mock_get_rpc_server.return_value = mock_server
        consumer = octavia_messaging_consumer.OctaviaConsumer(self.driver)
        consumer.start()
        consumer.stop()
        mock_server.stop.assert_called_once_with()
        mock_server.wait.assert_not_called()

    @mock.patch.object(octavia_messaging_consumer.messaging, 'get_rpc_server')
    def test_consumer_graceful_stop(self, mock_get_rpc_server):
        mock_server = mock.Mock()
        mock_get_rpc_server.return_value = mock_server
        consumer = octavia_messaging_consumer.OctaviaConsumer(self.driver)
        consumer.start()
        consumer.stop(graceful=True)
        mock_server.stop.assert_called_once_with()
        mock_server.wait.assert_called_once_with()

    @mock.patch.object(octavia_messaging_consumer.messaging, 'get_rpc_server')
    def test_consumer_reset(self, mock_get_rpc_server):
        mock_server = mock.Mock()
        mock_get_rpc_server.return_value = mock_server
        consumer = octavia_messaging_consumer.OctaviaConsumer(self.driver)
        consumer.start()
        consumer.reset()
        mock_server.reset.assert_called_once_with()

    def set_db_mocks(self):
        TOPIC = 'neutron_lbaas_event'
        self.addCleanup(cfg.CONF.clear_override, 'event_stream_topic',
                        group='oslo_messaging')
        cfg.CONF.set_override('event_stream_topic', TOPIC,
                              group='oslo_messaging')
        self.payload = {'operating_status': 'ONLINE'}
        self.consumer = octavia_messaging_consumer.OctaviaConsumer(
            self.driver)

    def test_updatedb_with_raises_exception_with_bad_model_name(self):
        self.set_db_mocks()

        cnt = InfoContainer('listener_statsX', 'id',
                            self.payload).to_dict()
        self.assertRaises(exceptions.ModelMapException,
                          self.consumer.endpoints[0].update_info, {}, cnt)

    def test_update_loadbalancer_stats(self):
        self.set_db_mocks()
        stats = {
            'bytes_in': 1,
            'bytes_out': 2,
            'active_connections': 3,
            'total_connections': 4,
            'request_errors': 5,
        }
        cnt = InfoContainer(constants.LOADBALANCER_STATS_EVENT, 'lb_id',
                            stats).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        self.driver.plugin.db.update_loadbalancer_stats.assert_called_with(
            mock.ANY, 'lb_id', stats)

    def test_updatedb_ignores_listener_stats(self):
        self.set_db_mocks()
        cnt = InfoContainer('listener_stats', 'id', self.payload).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        call_len = len(self.driver.plugin.db.update_status.call_args_list)
        self.assertEqual(call_len, 0)   # See didn't do anything

    def test_updatedb_loadbalancer(self):
        self.set_db_mocks()
        cnt = InfoContainer(constants.LOADBALANCER_EVENT, 'lb_id',
                            self.payload).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        self.assert_handle_streamed_event_called(models.LoadBalancer, 'lb_id',
                                                 self.payload)

    def test_updatedb_listener(self):
        self.set_db_mocks()
        cnt = InfoContainer(constants.LISTENER_EVENT, 'listener_id',
                            self.payload).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        self.assert_handle_streamed_event_called(models.Listener,
                                                 'listener_id',
                                                 self.payload)

    def test_updatedb_pool(self):
        self.set_db_mocks()
        cnt = InfoContainer(constants.POOL_EVENT, 'pool_id',
                            self.payload).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        self.assert_handle_streamed_event_called(models.PoolV2, 'pool_id',
                                                 self.payload)

    def test_updatedb_member(self):
        self.set_db_mocks()
        cnt = InfoContainer(constants.MEMBER_EVENT, 'pool_id',
                            self.payload).to_dict()
        self.consumer.endpoints[0].update_info({}, cnt)
        self.assert_handle_streamed_event_called(models.MemberV2, 'pool_id',
                                                 self.payload)
